# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, List, Union, Iterable, Optional
from datetime import date, datetime
from typing_extensions import Literal

import httpx

from .. import _legacy_response
from ..types import (
    invoice_list_params,
    invoice_issue_params,
    invoice_create_params,
    invoice_update_params,
    invoice_mark_paid_params,
    invoice_fetch_upcoming_params,
)
from .._types import Body, Omit, Query, Headers, NotGiven, omit, not_given
from .._utils import maybe_transform, async_maybe_transform
from .._compat import cached_property
from .._resource import SyncAPIResource, AsyncAPIResource
from .._response import to_streamed_response_wrapper, async_to_streamed_response_wrapper
from ..pagination import SyncPage, AsyncPage
from .._base_client import AsyncPaginator, make_request_options
from ..types.shared.invoice import Invoice
from ..types.shared_params.discount import Discount
from ..types.invoice_fetch_upcoming_response import InvoiceFetchUpcomingResponse

__all__ = ["Invoices", "AsyncInvoices"]


class Invoices(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> InvoicesWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/orbcorp/orb-python#accessing-raw-response-data-eg-headers
        """
        return InvoicesWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> InvoicesWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/orbcorp/orb-python#with_streaming_response
        """
        return InvoicesWithStreamingResponse(self)

    def create(
        self,
        *,
        currency: str,
        invoice_date: Union[str, datetime],
        line_items: Iterable[invoice_create_params.LineItem],
        customer_id: Optional[str] | Omit = omit,
        discount: Optional[Discount] | Omit = omit,
        due_date: Union[Union[str, date], Union[str, datetime], None] | Omit = omit,
        external_customer_id: Optional[str] | Omit = omit,
        memo: Optional[str] | Omit = omit,
        metadata: Optional[Dict[str, Optional[str]]] | Omit = omit,
        net_terms: Optional[int] | Omit = omit,
        will_auto_issue: bool | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> Invoice:
        """
        This endpoint is used to create a one-off invoice for a customer.

        Args:
          currency: An ISO 4217 currency string. Must be the same as the customer's currency if it
              is set.

          invoice_date: Optional invoice date to set. Must be in the past, if not set, `invoice_date` is
              set to the current time in the customer's timezone.

          customer_id: The id of the `Customer` to create this invoice for. One of `customer_id` and
              `external_customer_id` are required.

          discount: An optional discount to attach to the invoice.

          due_date: An optional custom due date for the invoice. If not set, the due date will be
              calculated based on the `net_terms` value.

          external_customer_id: The `external_customer_id` of the `Customer` to create this invoice for. One of
              `customer_id` and `external_customer_id` are required.

          memo: An optional memo to attach to the invoice. If no memo is provided, we will
              attach the default memo

          metadata: User-specified key/value pairs for the resource. Individual keys can be removed
              by setting the value to `null`, and the entire metadata mapping can be cleared
              by setting `metadata` to `null`.

          net_terms: The net terms determines the due date of the invoice. Due date is calculated
              based on the invoice or issuance date, depending on the account's configured due
              date calculation method. A value of '0' here represents that the invoice is due
              on issue, whereas a value of '30' represents that the customer has 30 days to
              pay the invoice. Do not set this field if you want to set a custom due date.

          will_auto_issue: When true, this invoice will be submitted for issuance upon creation. When
              false, the resulting invoice will require manual review to issue. Defaulted to
              false.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return self._post(
            "/invoices",
            body=maybe_transform(
                {
                    "currency": currency,
                    "invoice_date": invoice_date,
                    "line_items": line_items,
                    "customer_id": customer_id,
                    "discount": discount,
                    "due_date": due_date,
                    "external_customer_id": external_customer_id,
                    "memo": memo,
                    "metadata": metadata,
                    "net_terms": net_terms,
                    "will_auto_issue": will_auto_issue,
                },
                invoice_create_params.InvoiceCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=Invoice,
        )

    def update(
        self,
        invoice_id: str,
        *,
        due_date: Union[Union[str, date], Union[str, datetime], None] | Omit = omit,
        metadata: Optional[Dict[str, Optional[str]]] | Omit = omit,
        net_terms: Optional[int] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> Invoice:
        """
        This endpoint allows you to update the `metadata`, `net_terms`, and `due_date`
        properties on an invoice. If you pass null for the metadata value, it will clear
        any existing metadata for that invoice.

        `metadata` can be modified regardless of invoice state. `net_terms` and
        `due_date` can only be modified if the invoice is in a `draft` state.

        Args:
          due_date: An optional custom due date for the invoice. If not set, the due date will be
              calculated based on the `net_terms` value.

          metadata: User-specified key/value pairs for the resource. Individual keys can be removed
              by setting the value to `null`, and the entire metadata mapping can be cleared
              by setting `metadata` to `null`.

          net_terms: The net terms determines the due date of the invoice. Due date is calculated
              based on the invoice or issuance date, depending on the account's configured due
              date calculation method. A value of '0' here represents that the invoice is due
              on issue, whereas a value of '30' represents that the customer has 30 days to
              pay the invoice. Do not set this field if you want to set a custom due date.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        if not invoice_id:
            raise ValueError(f"Expected a non-empty value for `invoice_id` but received {invoice_id!r}")
        return self._put(
            f"/invoices/{invoice_id}",
            body=maybe_transform(
                {
                    "due_date": due_date,
                    "metadata": metadata,
                    "net_terms": net_terms,
                },
                invoice_update_params.InvoiceUpdateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=Invoice,
        )

    def list(
        self,
        *,
        amount: Optional[str] | Omit = omit,
        amount_gt: Optional[str] | Omit = omit,
        amount_lt: Optional[str] | Omit = omit,
        cursor: Optional[str] | Omit = omit,
        customer_id: Optional[str] | Omit = omit,
        date_type: Optional[Literal["due_date", "invoice_date"]] | Omit = omit,
        due_date: Union[str, date, None] | Omit = omit,
        due_date_window: Optional[str] | Omit = omit,
        due_date_gt: Union[str, date, None] | Omit = omit,
        due_date_lt: Union[str, date, None] | Omit = omit,
        external_customer_id: Optional[str] | Omit = omit,
        invoice_date_gt: Union[str, datetime, None] | Omit = omit,
        invoice_date_gte: Union[str, datetime, None] | Omit = omit,
        invoice_date_lt: Union[str, datetime, None] | Omit = omit,
        invoice_date_lte: Union[str, datetime, None] | Omit = omit,
        is_recurring: Optional[bool] | Omit = omit,
        limit: int | Omit = omit,
        status: Optional[List[Literal["draft", "issued", "paid", "synced", "void"]]] | Omit = omit,
        subscription_id: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SyncPage[Invoice]:
        """
        This endpoint returns a list of all [`Invoice`](/core-concepts#invoice)s for an
        account in a list format.

        The list of invoices is ordered starting from the most recently issued invoice
        date. The response also includes
        [`pagination_metadata`](/api-reference/pagination), which lets the caller
        retrieve the next page of results if they exist.

        By default, this only returns invoices that are `issued`, `paid`, or `synced`.

        When fetching any `draft` invoices, this returns the last-computed invoice
        values for each draft invoice, which may not always be up-to-date since Orb
        regularly refreshes invoices asynchronously.

        Args:
          cursor: Cursor for pagination. This can be populated by the `next_cursor` value returned
              from the initial request.

          due_date_window: Filters invoices by their due dates within a specific time range in the past.
              Specify the range as a number followed by 'd' (days) or 'm' (months). For
              example, '7d' filters invoices due in the last 7 days, and '2m' filters those
              due in the last 2 months.

          limit: The number of items to fetch. Defaults to 20.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get_api_list(
            "/invoices",
            page=SyncPage[Invoice],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "amount": amount,
                        "amount_gt": amount_gt,
                        "amount_lt": amount_lt,
                        "cursor": cursor,
                        "customer_id": customer_id,
                        "date_type": date_type,
                        "due_date": due_date,
                        "due_date_window": due_date_window,
                        "due_date_gt": due_date_gt,
                        "due_date_lt": due_date_lt,
                        "external_customer_id": external_customer_id,
                        "invoice_date_gt": invoice_date_gt,
                        "invoice_date_gte": invoice_date_gte,
                        "invoice_date_lt": invoice_date_lt,
                        "invoice_date_lte": invoice_date_lte,
                        "is_recurring": is_recurring,
                        "limit": limit,
                        "status": status,
                        "subscription_id": subscription_id,
                    },
                    invoice_list_params.InvoiceListParams,
                ),
            ),
            model=Invoice,
        )

    def fetch(
        self,
        invoice_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> Invoice:
        """
        This endpoint is used to fetch an [`Invoice`](/core-concepts#invoice) given an
        identifier.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not invoice_id:
            raise ValueError(f"Expected a non-empty value for `invoice_id` but received {invoice_id!r}")
        return self._get(
            f"/invoices/{invoice_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=Invoice,
        )

    def fetch_upcoming(
        self,
        *,
        subscription_id: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> InvoiceFetchUpcomingResponse:
        """
        This endpoint can be used to fetch the upcoming
        [invoice](/core-concepts#invoice) for the current billing period given a
        subscription.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get(
            "/invoices/upcoming",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {"subscription_id": subscription_id}, invoice_fetch_upcoming_params.InvoiceFetchUpcomingParams
                ),
            ),
            cast_to=InvoiceFetchUpcomingResponse,
        )

    def issue(
        self,
        invoice_id: str,
        *,
        synchronous: bool | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> Invoice:
        """This endpoint allows an eligible invoice to be issued manually.

        This is only
        possible with invoices where status is `draft`, `will_auto_issue` is false, and
        an `eligible_to_issue_at` is a time in the past. Issuing an invoice could
        possibly trigger side effects, some of which could be customer-visible (e.g.
        sending emails, auto-collecting payment, syncing the invoice to external
        providers, etc).

        Args:
          synchronous: If true, the invoice will be issued synchronously. If false, the invoice will be
              issued asynchronously. The synchronous option is only available for invoices
              that have no usage fees. If the invoice is configured to sync to an external
              provider, a successful response from this endpoint guarantees the invoice is
              present in the provider.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        if not invoice_id:
            raise ValueError(f"Expected a non-empty value for `invoice_id` but received {invoice_id!r}")
        return self._post(
            f"/invoices/{invoice_id}/issue",
            body=maybe_transform({"synchronous": synchronous}, invoice_issue_params.InvoiceIssueParams),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=Invoice,
        )

    def mark_paid(
        self,
        invoice_id: str,
        *,
        payment_received_date: Union[str, date],
        external_id: Optional[str] | Omit = omit,
        notes: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> Invoice:
        """This endpoint allows an invoice's status to be set to the `paid` status.

        This
        can only be done to invoices that are in the `issued` or `synced` status.

        Args:
          payment_received_date: A date string to specify the date of the payment.

          external_id: An optional external ID to associate with the payment.

          notes: An optional note to associate with the payment.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        if not invoice_id:
            raise ValueError(f"Expected a non-empty value for `invoice_id` but received {invoice_id!r}")
        return self._post(
            f"/invoices/{invoice_id}/mark_paid",
            body=maybe_transform(
                {
                    "payment_received_date": payment_received_date,
                    "external_id": external_id,
                    "notes": notes,
                },
                invoice_mark_paid_params.InvoiceMarkPaidParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=Invoice,
        )

    def pay(
        self,
        invoice_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> Invoice:
        """
        This endpoint collects payment for an invoice using the customer's default
        payment method. This action can only be taken on invoices with status "issued".

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        if not invoice_id:
            raise ValueError(f"Expected a non-empty value for `invoice_id` but received {invoice_id!r}")
        return self._post(
            f"/invoices/{invoice_id}/pay",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=Invoice,
        )

    def void(
        self,
        invoice_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> Invoice:
        """This endpoint allows an invoice's status to be set to the `void` status.

        This
        can only be done to invoices that are in the `issued` status.

        If the associated invoice has used the customer balance to change the amount
        due, the customer balance operation will be reverted. For example, if the
        invoice used \\$$10 of customer balance, that amount will be added back to the
        customer balance upon voiding.

        If the invoice was used to purchase a credit block, but the invoice is not yet
        paid, the credit block will be voided. If the invoice was created due to a
        top-up, the top-up will be disabled.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        if not invoice_id:
            raise ValueError(f"Expected a non-empty value for `invoice_id` but received {invoice_id!r}")
        return self._post(
            f"/invoices/{invoice_id}/void",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=Invoice,
        )


class AsyncInvoices(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncInvoicesWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/orbcorp/orb-python#accessing-raw-response-data-eg-headers
        """
        return AsyncInvoicesWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncInvoicesWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/orbcorp/orb-python#with_streaming_response
        """
        return AsyncInvoicesWithStreamingResponse(self)

    async def create(
        self,
        *,
        currency: str,
        invoice_date: Union[str, datetime],
        line_items: Iterable[invoice_create_params.LineItem],
        customer_id: Optional[str] | Omit = omit,
        discount: Optional[Discount] | Omit = omit,
        due_date: Union[Union[str, date], Union[str, datetime], None] | Omit = omit,
        external_customer_id: Optional[str] | Omit = omit,
        memo: Optional[str] | Omit = omit,
        metadata: Optional[Dict[str, Optional[str]]] | Omit = omit,
        net_terms: Optional[int] | Omit = omit,
        will_auto_issue: bool | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> Invoice:
        """
        This endpoint is used to create a one-off invoice for a customer.

        Args:
          currency: An ISO 4217 currency string. Must be the same as the customer's currency if it
              is set.

          invoice_date: Optional invoice date to set. Must be in the past, if not set, `invoice_date` is
              set to the current time in the customer's timezone.

          customer_id: The id of the `Customer` to create this invoice for. One of `customer_id` and
              `external_customer_id` are required.

          discount: An optional discount to attach to the invoice.

          due_date: An optional custom due date for the invoice. If not set, the due date will be
              calculated based on the `net_terms` value.

          external_customer_id: The `external_customer_id` of the `Customer` to create this invoice for. One of
              `customer_id` and `external_customer_id` are required.

          memo: An optional memo to attach to the invoice. If no memo is provided, we will
              attach the default memo

          metadata: User-specified key/value pairs for the resource. Individual keys can be removed
              by setting the value to `null`, and the entire metadata mapping can be cleared
              by setting `metadata` to `null`.

          net_terms: The net terms determines the due date of the invoice. Due date is calculated
              based on the invoice or issuance date, depending on the account's configured due
              date calculation method. A value of '0' here represents that the invoice is due
              on issue, whereas a value of '30' represents that the customer has 30 days to
              pay the invoice. Do not set this field if you want to set a custom due date.

          will_auto_issue: When true, this invoice will be submitted for issuance upon creation. When
              false, the resulting invoice will require manual review to issue. Defaulted to
              false.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return await self._post(
            "/invoices",
            body=await async_maybe_transform(
                {
                    "currency": currency,
                    "invoice_date": invoice_date,
                    "line_items": line_items,
                    "customer_id": customer_id,
                    "discount": discount,
                    "due_date": due_date,
                    "external_customer_id": external_customer_id,
                    "memo": memo,
                    "metadata": metadata,
                    "net_terms": net_terms,
                    "will_auto_issue": will_auto_issue,
                },
                invoice_create_params.InvoiceCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=Invoice,
        )

    async def update(
        self,
        invoice_id: str,
        *,
        due_date: Union[Union[str, date], Union[str, datetime], None] | Omit = omit,
        metadata: Optional[Dict[str, Optional[str]]] | Omit = omit,
        net_terms: Optional[int] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> Invoice:
        """
        This endpoint allows you to update the `metadata`, `net_terms`, and `due_date`
        properties on an invoice. If you pass null for the metadata value, it will clear
        any existing metadata for that invoice.

        `metadata` can be modified regardless of invoice state. `net_terms` and
        `due_date` can only be modified if the invoice is in a `draft` state.

        Args:
          due_date: An optional custom due date for the invoice. If not set, the due date will be
              calculated based on the `net_terms` value.

          metadata: User-specified key/value pairs for the resource. Individual keys can be removed
              by setting the value to `null`, and the entire metadata mapping can be cleared
              by setting `metadata` to `null`.

          net_terms: The net terms determines the due date of the invoice. Due date is calculated
              based on the invoice or issuance date, depending on the account's configured due
              date calculation method. A value of '0' here represents that the invoice is due
              on issue, whereas a value of '30' represents that the customer has 30 days to
              pay the invoice. Do not set this field if you want to set a custom due date.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        if not invoice_id:
            raise ValueError(f"Expected a non-empty value for `invoice_id` but received {invoice_id!r}")
        return await self._put(
            f"/invoices/{invoice_id}",
            body=await async_maybe_transform(
                {
                    "due_date": due_date,
                    "metadata": metadata,
                    "net_terms": net_terms,
                },
                invoice_update_params.InvoiceUpdateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=Invoice,
        )

    def list(
        self,
        *,
        amount: Optional[str] | Omit = omit,
        amount_gt: Optional[str] | Omit = omit,
        amount_lt: Optional[str] | Omit = omit,
        cursor: Optional[str] | Omit = omit,
        customer_id: Optional[str] | Omit = omit,
        date_type: Optional[Literal["due_date", "invoice_date"]] | Omit = omit,
        due_date: Union[str, date, None] | Omit = omit,
        due_date_window: Optional[str] | Omit = omit,
        due_date_gt: Union[str, date, None] | Omit = omit,
        due_date_lt: Union[str, date, None] | Omit = omit,
        external_customer_id: Optional[str] | Omit = omit,
        invoice_date_gt: Union[str, datetime, None] | Omit = omit,
        invoice_date_gte: Union[str, datetime, None] | Omit = omit,
        invoice_date_lt: Union[str, datetime, None] | Omit = omit,
        invoice_date_lte: Union[str, datetime, None] | Omit = omit,
        is_recurring: Optional[bool] | Omit = omit,
        limit: int | Omit = omit,
        status: Optional[List[Literal["draft", "issued", "paid", "synced", "void"]]] | Omit = omit,
        subscription_id: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> AsyncPaginator[Invoice, AsyncPage[Invoice]]:
        """
        This endpoint returns a list of all [`Invoice`](/core-concepts#invoice)s for an
        account in a list format.

        The list of invoices is ordered starting from the most recently issued invoice
        date. The response also includes
        [`pagination_metadata`](/api-reference/pagination), which lets the caller
        retrieve the next page of results if they exist.

        By default, this only returns invoices that are `issued`, `paid`, or `synced`.

        When fetching any `draft` invoices, this returns the last-computed invoice
        values for each draft invoice, which may not always be up-to-date since Orb
        regularly refreshes invoices asynchronously.

        Args:
          cursor: Cursor for pagination. This can be populated by the `next_cursor` value returned
              from the initial request.

          due_date_window: Filters invoices by their due dates within a specific time range in the past.
              Specify the range as a number followed by 'd' (days) or 'm' (months). For
              example, '7d' filters invoices due in the last 7 days, and '2m' filters those
              due in the last 2 months.

          limit: The number of items to fetch. Defaults to 20.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get_api_list(
            "/invoices",
            page=AsyncPage[Invoice],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "amount": amount,
                        "amount_gt": amount_gt,
                        "amount_lt": amount_lt,
                        "cursor": cursor,
                        "customer_id": customer_id,
                        "date_type": date_type,
                        "due_date": due_date,
                        "due_date_window": due_date_window,
                        "due_date_gt": due_date_gt,
                        "due_date_lt": due_date_lt,
                        "external_customer_id": external_customer_id,
                        "invoice_date_gt": invoice_date_gt,
                        "invoice_date_gte": invoice_date_gte,
                        "invoice_date_lt": invoice_date_lt,
                        "invoice_date_lte": invoice_date_lte,
                        "is_recurring": is_recurring,
                        "limit": limit,
                        "status": status,
                        "subscription_id": subscription_id,
                    },
                    invoice_list_params.InvoiceListParams,
                ),
            ),
            model=Invoice,
        )

    async def fetch(
        self,
        invoice_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> Invoice:
        """
        This endpoint is used to fetch an [`Invoice`](/core-concepts#invoice) given an
        identifier.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not invoice_id:
            raise ValueError(f"Expected a non-empty value for `invoice_id` but received {invoice_id!r}")
        return await self._get(
            f"/invoices/{invoice_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=Invoice,
        )

    async def fetch_upcoming(
        self,
        *,
        subscription_id: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> InvoiceFetchUpcomingResponse:
        """
        This endpoint can be used to fetch the upcoming
        [invoice](/core-concepts#invoice) for the current billing period given a
        subscription.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._get(
            "/invoices/upcoming",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=await async_maybe_transform(
                    {"subscription_id": subscription_id}, invoice_fetch_upcoming_params.InvoiceFetchUpcomingParams
                ),
            ),
            cast_to=InvoiceFetchUpcomingResponse,
        )

    async def issue(
        self,
        invoice_id: str,
        *,
        synchronous: bool | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> Invoice:
        """This endpoint allows an eligible invoice to be issued manually.

        This is only
        possible with invoices where status is `draft`, `will_auto_issue` is false, and
        an `eligible_to_issue_at` is a time in the past. Issuing an invoice could
        possibly trigger side effects, some of which could be customer-visible (e.g.
        sending emails, auto-collecting payment, syncing the invoice to external
        providers, etc).

        Args:
          synchronous: If true, the invoice will be issued synchronously. If false, the invoice will be
              issued asynchronously. The synchronous option is only available for invoices
              that have no usage fees. If the invoice is configured to sync to an external
              provider, a successful response from this endpoint guarantees the invoice is
              present in the provider.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        if not invoice_id:
            raise ValueError(f"Expected a non-empty value for `invoice_id` but received {invoice_id!r}")
        return await self._post(
            f"/invoices/{invoice_id}/issue",
            body=await async_maybe_transform({"synchronous": synchronous}, invoice_issue_params.InvoiceIssueParams),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=Invoice,
        )

    async def mark_paid(
        self,
        invoice_id: str,
        *,
        payment_received_date: Union[str, date],
        external_id: Optional[str] | Omit = omit,
        notes: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> Invoice:
        """This endpoint allows an invoice's status to be set to the `paid` status.

        This
        can only be done to invoices that are in the `issued` or `synced` status.

        Args:
          payment_received_date: A date string to specify the date of the payment.

          external_id: An optional external ID to associate with the payment.

          notes: An optional note to associate with the payment.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        if not invoice_id:
            raise ValueError(f"Expected a non-empty value for `invoice_id` but received {invoice_id!r}")
        return await self._post(
            f"/invoices/{invoice_id}/mark_paid",
            body=await async_maybe_transform(
                {
                    "payment_received_date": payment_received_date,
                    "external_id": external_id,
                    "notes": notes,
                },
                invoice_mark_paid_params.InvoiceMarkPaidParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=Invoice,
        )

    async def pay(
        self,
        invoice_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> Invoice:
        """
        This endpoint collects payment for an invoice using the customer's default
        payment method. This action can only be taken on invoices with status "issued".

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        if not invoice_id:
            raise ValueError(f"Expected a non-empty value for `invoice_id` but received {invoice_id!r}")
        return await self._post(
            f"/invoices/{invoice_id}/pay",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=Invoice,
        )

    async def void(
        self,
        invoice_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> Invoice:
        """This endpoint allows an invoice's status to be set to the `void` status.

        This
        can only be done to invoices that are in the `issued` status.

        If the associated invoice has used the customer balance to change the amount
        due, the customer balance operation will be reverted. For example, if the
        invoice used \\$$10 of customer balance, that amount will be added back to the
        customer balance upon voiding.

        If the invoice was used to purchase a credit block, but the invoice is not yet
        paid, the credit block will be voided. If the invoice was created due to a
        top-up, the top-up will be disabled.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        if not invoice_id:
            raise ValueError(f"Expected a non-empty value for `invoice_id` but received {invoice_id!r}")
        return await self._post(
            f"/invoices/{invoice_id}/void",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=Invoice,
        )


class InvoicesWithRawResponse:
    def __init__(self, invoices: Invoices) -> None:
        self._invoices = invoices

        self.create = _legacy_response.to_raw_response_wrapper(
            invoices.create,
        )
        self.update = _legacy_response.to_raw_response_wrapper(
            invoices.update,
        )
        self.list = _legacy_response.to_raw_response_wrapper(
            invoices.list,
        )
        self.fetch = _legacy_response.to_raw_response_wrapper(
            invoices.fetch,
        )
        self.fetch_upcoming = _legacy_response.to_raw_response_wrapper(
            invoices.fetch_upcoming,
        )
        self.issue = _legacy_response.to_raw_response_wrapper(
            invoices.issue,
        )
        self.mark_paid = _legacy_response.to_raw_response_wrapper(
            invoices.mark_paid,
        )
        self.pay = _legacy_response.to_raw_response_wrapper(
            invoices.pay,
        )
        self.void = _legacy_response.to_raw_response_wrapper(
            invoices.void,
        )


class AsyncInvoicesWithRawResponse:
    def __init__(self, invoices: AsyncInvoices) -> None:
        self._invoices = invoices

        self.create = _legacy_response.async_to_raw_response_wrapper(
            invoices.create,
        )
        self.update = _legacy_response.async_to_raw_response_wrapper(
            invoices.update,
        )
        self.list = _legacy_response.async_to_raw_response_wrapper(
            invoices.list,
        )
        self.fetch = _legacy_response.async_to_raw_response_wrapper(
            invoices.fetch,
        )
        self.fetch_upcoming = _legacy_response.async_to_raw_response_wrapper(
            invoices.fetch_upcoming,
        )
        self.issue = _legacy_response.async_to_raw_response_wrapper(
            invoices.issue,
        )
        self.mark_paid = _legacy_response.async_to_raw_response_wrapper(
            invoices.mark_paid,
        )
        self.pay = _legacy_response.async_to_raw_response_wrapper(
            invoices.pay,
        )
        self.void = _legacy_response.async_to_raw_response_wrapper(
            invoices.void,
        )


class InvoicesWithStreamingResponse:
    def __init__(self, invoices: Invoices) -> None:
        self._invoices = invoices

        self.create = to_streamed_response_wrapper(
            invoices.create,
        )
        self.update = to_streamed_response_wrapper(
            invoices.update,
        )
        self.list = to_streamed_response_wrapper(
            invoices.list,
        )
        self.fetch = to_streamed_response_wrapper(
            invoices.fetch,
        )
        self.fetch_upcoming = to_streamed_response_wrapper(
            invoices.fetch_upcoming,
        )
        self.issue = to_streamed_response_wrapper(
            invoices.issue,
        )
        self.mark_paid = to_streamed_response_wrapper(
            invoices.mark_paid,
        )
        self.pay = to_streamed_response_wrapper(
            invoices.pay,
        )
        self.void = to_streamed_response_wrapper(
            invoices.void,
        )


class AsyncInvoicesWithStreamingResponse:
    def __init__(self, invoices: AsyncInvoices) -> None:
        self._invoices = invoices

        self.create = async_to_streamed_response_wrapper(
            invoices.create,
        )
        self.update = async_to_streamed_response_wrapper(
            invoices.update,
        )
        self.list = async_to_streamed_response_wrapper(
            invoices.list,
        )
        self.fetch = async_to_streamed_response_wrapper(
            invoices.fetch,
        )
        self.fetch_upcoming = async_to_streamed_response_wrapper(
            invoices.fetch_upcoming,
        )
        self.issue = async_to_streamed_response_wrapper(
            invoices.issue,
        )
        self.mark_paid = async_to_streamed_response_wrapper(
            invoices.mark_paid,
        )
        self.pay = async_to_streamed_response_wrapper(
            invoices.pay,
        )
        self.void = async_to_streamed_response_wrapper(
            invoices.void,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Optional
from typing_extensions import Literal, Required, TypedDict

__all__ = ["DiscountOverrideParam"]


class DiscountOverrideParam(TypedDict, total=False):
    discount_type: Required[Literal["percentage", "usage", "amount"]]

    amount_discount: Optional[str]
    """Only available if discount_type is `amount`."""

    percentage_discount: Optional[float]
    """Only available if discount_type is `percentage`.

    This is a number between 0 and 1.
    """

    usage_discount: Optional[float]
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional
from datetime import datetime
from typing_extensions import Literal

from ..._models import BaseModel

__all__ = ["BackfillRevertResponse"]


class BackfillRevertResponse(BaseModel):
    id: str

    close_time: Optional[datetime] = None
    """If in the future, the time at which the backfill will automatically close.

    If in the past, the time at which the backfill was closed.
    """

    created_at: datetime

    customer_id: Optional[str] = None
    """The Orb-generated ID of the customer to which this backfill is scoped.

    If `null`, this backfill is scoped to all customers.
    """

    events_ingested: int
    """The number of events ingested in this backfill."""

    replace_existing_events: bool
    """
    If `true`, existing events in the backfill's timeframe will be replaced with the
    newly ingested events associated with the backfill. If `false`, newly ingested
    events will be added to the existing events.
    """

    reverted_at: Optional[datetime] = None
    """The time at which this backfill was reverted."""

    status: Literal["pending", "reflected", "pending_revert", "reverted"]
    """The status of the backfill."""

    timeframe_end: datetime

    timeframe_start: datetime

    deprecation_filter: Optional[str] = None
    """
    A boolean
    [computed property](/extensibility/advanced-metrics#computed-properties) used to
    filter the set of events to deprecate
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Optional

import httpx

from ... import _legacy_response
from ...types import (
    dimensional_price_group_list_params,
    dimensional_price_group_create_params,
    dimensional_price_group_update_params,
)
from ..._types import Body, Omit, Query, Headers, NotGiven, SequenceNotStr, omit, not_given
from ..._utils import maybe_transform, async_maybe_transform
from ..._compat import cached_property
from ..._resource import SyncAPIResource, AsyncAPIResource
from ..._response import to_streamed_response_wrapper, async_to_streamed_response_wrapper
from ...pagination import SyncPage, AsyncPage
from ..._base_client import AsyncPaginator, make_request_options
from ...types.dimensional_price_group import DimensionalPriceGroup
from .external_dimensional_price_group_id import (
    ExternalDimensionalPriceGroupID,
    AsyncExternalDimensionalPriceGroupID,
    ExternalDimensionalPriceGroupIDWithRawResponse,
    AsyncExternalDimensionalPriceGroupIDWithRawResponse,
    ExternalDimensionalPriceGroupIDWithStreamingResponse,
    AsyncExternalDimensionalPriceGroupIDWithStreamingResponse,
)

__all__ = ["DimensionalPriceGroups", "AsyncDimensionalPriceGroups"]


class DimensionalPriceGroups(SyncAPIResource):
    @cached_property
    def external_dimensional_price_group_id(self) -> ExternalDimensionalPriceGroupID:
        return ExternalDimensionalPriceGroupID(self._client)

    @cached_property
    def with_raw_response(self) -> DimensionalPriceGroupsWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/orbcorp/orb-python#accessing-raw-response-data-eg-headers
        """
        return DimensionalPriceGroupsWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> DimensionalPriceGroupsWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/orbcorp/orb-python#with_streaming_response
        """
        return DimensionalPriceGroupsWithStreamingResponse(self)

    def create(
        self,
        *,
        billable_metric_id: str,
        dimensions: SequenceNotStr[str],
        name: str,
        external_dimensional_price_group_id: Optional[str] | Omit = omit,
        metadata: Optional[Dict[str, Optional[str]]] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> DimensionalPriceGroup:
        """
        A dimensional price group is used to partition the result of a billable metric
        by a set of dimensions. Prices in a price group must specify the partition used
        to derive their usage.

        For example, suppose we have a billable metric that measures the number of
        widgets used and we want to charge differently depending on the color of the
        widget. We can create a price group with a dimension "color" and two prices: one
        that charges \\$$10 per red widget and one that charges \\$$20 per blue widget.

        Args:
          dimensions: The set of keys (in order) used to disambiguate prices in the group.

          metadata: User-specified key/value pairs for the resource. Individual keys can be removed
              by setting the value to `null`, and the entire metadata mapping can be cleared
              by setting `metadata` to `null`.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return self._post(
            "/dimensional_price_groups",
            body=maybe_transform(
                {
                    "billable_metric_id": billable_metric_id,
                    "dimensions": dimensions,
                    "name": name,
                    "external_dimensional_price_group_id": external_dimensional_price_group_id,
                    "metadata": metadata,
                },
                dimensional_price_group_create_params.DimensionalPriceGroupCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=DimensionalPriceGroup,
        )

    def retrieve(
        self,
        dimensional_price_group_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> DimensionalPriceGroup:
        """
        Fetch dimensional price group

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not dimensional_price_group_id:
            raise ValueError(
                f"Expected a non-empty value for `dimensional_price_group_id` but received {dimensional_price_group_id!r}"
            )
        return self._get(
            f"/dimensional_price_groups/{dimensional_price_group_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=DimensionalPriceGroup,
        )

    def update(
        self,
        dimensional_price_group_id: str,
        *,
        external_dimensional_price_group_id: Optional[str] | Omit = omit,
        metadata: Optional[Dict[str, Optional[str]]] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> DimensionalPriceGroup:
        """
        This endpoint can be used to update the `external_dimensional_price_group_id`
        and `metadata` of an existing dimensional price group. Other fields on a
        dimensional price group are currently immutable.

        Args:
          external_dimensional_price_group_id: An optional user-defined ID for this dimensional price group resource, used
              throughout the system as an alias for this dimensional price group. Use this
              field to identify a dimensional price group by an existing identifier in your
              system.

          metadata: User-specified key/value pairs for the resource. Individual keys can be removed
              by setting the value to `null`, and the entire metadata mapping can be cleared
              by setting `metadata` to `null`.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        if not dimensional_price_group_id:
            raise ValueError(
                f"Expected a non-empty value for `dimensional_price_group_id` but received {dimensional_price_group_id!r}"
            )
        return self._put(
            f"/dimensional_price_groups/{dimensional_price_group_id}",
            body=maybe_transform(
                {
                    "external_dimensional_price_group_id": external_dimensional_price_group_id,
                    "metadata": metadata,
                },
                dimensional_price_group_update_params.DimensionalPriceGroupUpdateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=DimensionalPriceGroup,
        )

    def list(
        self,
        *,
        cursor: Optional[str] | Omit = omit,
        limit: int | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SyncPage[DimensionalPriceGroup]:
        """List dimensional price groups

        Args:
          cursor: Cursor for pagination.

        This can be populated by the `next_cursor` value returned
              from the initial request.

          limit: The number of items to fetch. Defaults to 20.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get_api_list(
            "/dimensional_price_groups",
            page=SyncPage[DimensionalPriceGroup],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "cursor": cursor,
                        "limit": limit,
                    },
                    dimensional_price_group_list_params.DimensionalPriceGroupListParams,
                ),
            ),
            model=DimensionalPriceGroup,
        )


class AsyncDimensionalPriceGroups(AsyncAPIResource):
    @cached_property
    def external_dimensional_price_group_id(self) -> AsyncExternalDimensionalPriceGroupID:
        return AsyncExternalDimensionalPriceGroupID(self._client)

    @cached_property
    def with_raw_response(self) -> AsyncDimensionalPriceGroupsWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/orbcorp/orb-python#accessing-raw-response-data-eg-headers
        """
        return AsyncDimensionalPriceGroupsWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncDimensionalPriceGroupsWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/orbcorp/orb-python#with_streaming_response
        """
        return AsyncDimensionalPriceGroupsWithStreamingResponse(self)

    async def create(
        self,
        *,
        billable_metric_id: str,
        dimensions: SequenceNotStr[str],
        name: str,
        external_dimensional_price_group_id: Optional[str] | Omit = omit,
        metadata: Optional[Dict[str, Optional[str]]] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> DimensionalPriceGroup:
        """
        A dimensional price group is used to partition the result of a billable metric
        by a set of dimensions. Prices in a price group must specify the partition used
        to derive their usage.

        For example, suppose we have a billable metric that measures the number of
        widgets used and we want to charge differently depending on the color of the
        widget. We can create a price group with a dimension "color" and two prices: one
        that charges \\$$10 per red widget and one that charges \\$$20 per blue widget.

        Args:
          dimensions: The set of keys (in order) used to disambiguate prices in the group.

          metadata: User-specified key/value pairs for the resource. Individual keys can be removed
              by setting the value to `null`, and the entire metadata mapping can be cleared
              by setting `metadata` to `null`.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return await self._post(
            "/dimensional_price_groups",
            body=await async_maybe_transform(
                {
                    "billable_metric_id": billable_metric_id,
                    "dimensions": dimensions,
                    "name": name,
                    "external_dimensional_price_group_id": external_dimensional_price_group_id,
                    "metadata": metadata,
                },
                dimensional_price_group_create_params.DimensionalPriceGroupCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=DimensionalPriceGroup,
        )

    async def retrieve(
        self,
        dimensional_price_group_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> DimensionalPriceGroup:
        """
        Fetch dimensional price group

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not dimensional_price_group_id:
            raise ValueError(
                f"Expected a non-empty value for `dimensional_price_group_id` but received {dimensional_price_group_id!r}"
            )
        return await self._get(
            f"/dimensional_price_groups/{dimensional_price_group_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=DimensionalPriceGroup,
        )

    async def update(
        self,
        dimensional_price_group_id: str,
        *,
        external_dimensional_price_group_id: Optional[str] | Omit = omit,
        metadata: Optional[Dict[str, Optional[str]]] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> DimensionalPriceGroup:
        """
        This endpoint can be used to update the `external_dimensional_price_group_id`
        and `metadata` of an existing dimensional price group. Other fields on a
        dimensional price group are currently immutable.

        Args:
          external_dimensional_price_group_id: An optional user-defined ID for this dimensional price group resource, used
              throughout the system as an alias for this dimensional price group. Use this
              field to identify a dimensional price group by an existing identifier in your
              system.

          metadata: User-specified key/value pairs for the resource. Individual keys can be removed
              by setting the value to `null`, and the entire metadata mapping can be cleared
              by setting `metadata` to `null`.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        if not dimensional_price_group_id:
            raise ValueError(
                f"Expected a non-empty value for `dimensional_price_group_id` but received {dimensional_price_group_id!r}"
            )
        return await self._put(
            f"/dimensional_price_groups/{dimensional_price_group_id}",
            body=await async_maybe_transform(
                {
                    "external_dimensional_price_group_id": external_dimensional_price_group_id,
                    "metadata": metadata,
                },
                dimensional_price_group_update_params.DimensionalPriceGroupUpdateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=DimensionalPriceGroup,
        )

    def list(
        self,
        *,
        cursor: Optional[str] | Omit = omit,
        limit: int | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> AsyncPaginator[DimensionalPriceGroup, AsyncPage[DimensionalPriceGroup]]:
        """List dimensional price groups

        Args:
          cursor: Cursor for pagination.

        This can be populated by the `next_cursor` value returned
              from the initial request.

          limit: The number of items to fetch. Defaults to 20.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get_api_list(
            "/dimensional_price_groups",
            page=AsyncPage[DimensionalPriceGroup],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "cursor": cursor,
                        "limit": limit,
                    },
                    dimensional_price_group_list_params.DimensionalPriceGroupListParams,
                ),
            ),
            model=DimensionalPriceGroup,
        )


class DimensionalPriceGroupsWithRawResponse:
    def __init__(self, dimensional_price_groups: DimensionalPriceGroups) -> None:
        self._dimensional_price_groups = dimensional_price_groups

        self.create = _legacy_response.to_raw_response_wrapper(
            dimensional_price_groups.create,
        )
        self.retrieve = _legacy_response.to_raw_response_wrapper(
            dimensional_price_groups.retrieve,
        )
        self.update = _legacy_response.to_raw_response_wrapper(
            dimensional_price_groups.update,
        )
        self.list = _legacy_response.to_raw_response_wrapper(
            dimensional_price_groups.list,
        )

    @cached_property
    def external_dimensional_price_group_id(self) -> ExternalDimensionalPriceGroupIDWithRawResponse:
        return ExternalDimensionalPriceGroupIDWithRawResponse(
            self._dimensional_price_groups.external_dimensional_price_group_id
        )


class AsyncDimensionalPriceGroupsWithRawResponse:
    def __init__(self, dimensional_price_groups: AsyncDimensionalPriceGroups) -> None:
        self._dimensional_price_groups = dimensional_price_groups

        self.create = _legacy_response.async_to_raw_response_wrapper(
            dimensional_price_groups.create,
        )
        self.retrieve = _legacy_response.async_to_raw_response_wrapper(
            dimensional_price_groups.retrieve,
        )
        self.update = _legacy_response.async_to_raw_response_wrapper(
            dimensional_price_groups.update,
        )
        self.list = _legacy_response.async_to_raw_response_wrapper(
            dimensional_price_groups.list,
        )

    @cached_property
    def external_dimensional_price_group_id(self) -> AsyncExternalDimensionalPriceGroupIDWithRawResponse:
        return AsyncExternalDimensionalPriceGroupIDWithRawResponse(
            self._dimensional_price_groups.external_dimensional_price_group_id
        )


class DimensionalPriceGroupsWithStreamingResponse:
    def __init__(self, dimensional_price_groups: DimensionalPriceGroups) -> None:
        self._dimensional_price_groups = dimensional_price_groups

        self.create = to_streamed_response_wrapper(
            dimensional_price_groups.create,
        )
        self.retrieve = to_streamed_response_wrapper(
            dimensional_price_groups.retrieve,
        )
        self.update = to_streamed_response_wrapper(
            dimensional_price_groups.update,
        )
        self.list = to_streamed_response_wrapper(
            dimensional_price_groups.list,
        )

    @cached_property
    def external_dimensional_price_group_id(self) -> ExternalDimensionalPriceGroupIDWithStreamingResponse:
        return ExternalDimensionalPriceGroupIDWithStreamingResponse(
            self._dimensional_price_groups.external_dimensional_price_group_id
        )


class AsyncDimensionalPriceGroupsWithStreamingResponse:
    def __init__(self, dimensional_price_groups: AsyncDimensionalPriceGroups) -> None:
        self._dimensional_price_groups = dimensional_price_groups

        self.create = async_to_streamed_response_wrapper(
            dimensional_price_groups.create,
        )
        self.retrieve = async_to_streamed_response_wrapper(
            dimensional_price_groups.retrieve,
        )
        self.update = async_to_streamed_response_wrapper(
            dimensional_price_groups.update,
        )
        self.list = async_to_streamed_response_wrapper(
            dimensional_price_groups.list,
        )

    @cached_property
    def external_dimensional_price_group_id(self) -> AsyncExternalDimensionalPriceGroupIDWithStreamingResponse:
        return AsyncExternalDimensionalPriceGroupIDWithStreamingResponse(
            self._dimensional_price_groups.external_dimensional_price_group_id
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional
from datetime import datetime
from typing_extensions import Literal

from ..._models import BaseModel
from .transform_price_filter import TransformPriceFilter

__all__ = ["UsageDiscountInterval"]


class UsageDiscountInterval(BaseModel):
    applies_to_price_interval_ids: List[str]
    """The price interval ids that this discount interval applies to."""

    discount_type: Literal["usage"]

    end_date: Optional[datetime] = None
    """The end date of the discount interval."""

    filters: List[TransformPriceFilter]
    """The filters that determine which prices this discount interval applies to."""

    start_date: datetime
    """The start date of the discount interval."""

    usage_discount: float
    """Only available if discount_type is `usage`.

    Number of usage units that this discount is for
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Any, Dict, Optional, cast

import httpx

from ... import _legacy_response
from ..._types import Body, Omit, Query, Headers, NotGiven, omit, not_given
from ..._utils import maybe_transform, async_maybe_transform
from ..._compat import cached_property
from ..._resource import SyncAPIResource, AsyncAPIResource
from ..._response import to_streamed_response_wrapper, async_to_streamed_response_wrapper
from ..._base_client import make_request_options
from ...types.prices import external_price_id_update_params
from ...types.shared.price import Price

__all__ = ["ExternalPriceID", "AsyncExternalPriceID"]


class ExternalPriceID(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> ExternalPriceIDWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/orbcorp/orb-python#accessing-raw-response-data-eg-headers
        """
        return ExternalPriceIDWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> ExternalPriceIDWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/orbcorp/orb-python#with_streaming_response
        """
        return ExternalPriceIDWithStreamingResponse(self)

    def update(
        self,
        external_price_id: str,
        *,
        metadata: Optional[Dict[str, Optional[str]]] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> Price:
        """This endpoint allows you to update the `metadata` property on a price.

        If you
        pass null for the metadata value, it will clear any existing metadata for that
        price.

        Args:
          metadata: User-specified key/value pairs for the resource. Individual keys can be removed
              by setting the value to `null`, and the entire metadata mapping can be cleared
              by setting `metadata` to `null`.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        if not external_price_id:
            raise ValueError(f"Expected a non-empty value for `external_price_id` but received {external_price_id!r}")
        return cast(
            Price,
            self._put(
                f"/prices/external_price_id/{external_price_id}",
                body=maybe_transform(
                    {"metadata": metadata}, external_price_id_update_params.ExternalPriceIDUpdateParams
                ),
                options=make_request_options(
                    extra_headers=extra_headers,
                    extra_query=extra_query,
                    extra_body=extra_body,
                    timeout=timeout,
                    idempotency_key=idempotency_key,
                ),
                cast_to=cast(Any, Price),  # Union types cannot be passed in as arguments in the type system
            ),
        )

    def fetch(
        self,
        external_price_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> Price:
        """This endpoint returns a price given an external price id.

        See the
        [price creation API](/api-reference/price/create-price) for more information
        about external price aliases.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not external_price_id:
            raise ValueError(f"Expected a non-empty value for `external_price_id` but received {external_price_id!r}")
        return cast(
            Price,
            self._get(
                f"/prices/external_price_id/{external_price_id}",
                options=make_request_options(
                    extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
                ),
                cast_to=cast(Any, Price),  # Union types cannot be passed in as arguments in the type system
            ),
        )


class AsyncExternalPriceID(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncExternalPriceIDWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/orbcorp/orb-python#accessing-raw-response-data-eg-headers
        """
        return AsyncExternalPriceIDWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncExternalPriceIDWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/orbcorp/orb-python#with_streaming_response
        """
        return AsyncExternalPriceIDWithStreamingResponse(self)

    async def update(
        self,
        external_price_id: str,
        *,
        metadata: Optional[Dict[str, Optional[str]]] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> Price:
        """This endpoint allows you to update the `metadata` property on a price.

        If you
        pass null for the metadata value, it will clear any existing metadata for that
        price.

        Args:
          metadata: User-specified key/value pairs for the resource. Individual keys can be removed
              by setting the value to `null`, and the entire metadata mapping can be cleared
              by setting `metadata` to `null`.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        if not external_price_id:
            raise ValueError(f"Expected a non-empty value for `external_price_id` but received {external_price_id!r}")
        return cast(
            Price,
            await self._put(
                f"/prices/external_price_id/{external_price_id}",
                body=await async_maybe_transform(
                    {"metadata": metadata}, external_price_id_update_params.ExternalPriceIDUpdateParams
                ),
                options=make_request_options(
                    extra_headers=extra_headers,
                    extra_query=extra_query,
                    extra_body=extra_body,
                    timeout=timeout,
                    idempotency_key=idempotency_key,
                ),
                cast_to=cast(Any, Price),  # Union types cannot be passed in as arguments in the type system
            ),
        )

    async def fetch(
        self,
        external_price_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> Price:
        """This endpoint returns a price given an external price id.

        See the
        [price creation API](/api-reference/price/create-price) for more information
        about external price aliases.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not external_price_id:
            raise ValueError(f"Expected a non-empty value for `external_price_id` but received {external_price_id!r}")
        return cast(
            Price,
            await self._get(
                f"/prices/external_price_id/{external_price_id}",
                options=make_request_options(
                    extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
                ),
                cast_to=cast(Any, Price),  # Union types cannot be passed in as arguments in the type system
            ),
        )


class ExternalPriceIDWithRawResponse:
    def __init__(self, external_price_id: ExternalPriceID) -> None:
        self._external_price_id = external_price_id

        self.update = _legacy_response.to_raw_response_wrapper(
            external_price_id.update,
        )
        self.fetch = _legacy_response.to_raw_response_wrapper(
            external_price_id.fetch,
        )


class AsyncExternalPriceIDWithRawResponse:
    def __init__(self, external_price_id: AsyncExternalPriceID) -> None:
        self._external_price_id = external_price_id

        self.update = _legacy_response.async_to_raw_response_wrapper(
            external_price_id.update,
        )
        self.fetch = _legacy_response.async_to_raw_response_wrapper(
            external_price_id.fetch,
        )


class ExternalPriceIDWithStreamingResponse:
    def __init__(self, external_price_id: ExternalPriceID) -> None:
        self._external_price_id = external_price_id

        self.update = to_streamed_response_wrapper(
            external_price_id.update,
        )
        self.fetch = to_streamed_response_wrapper(
            external_price_id.fetch,
        )


class AsyncExternalPriceIDWithStreamingResponse:
    def __init__(self, external_price_id: AsyncExternalPriceID) -> None:
        self._external_price_id = external_price_id

        self.update = async_to_streamed_response_wrapper(
            external_price_id.update,
        )
        self.fetch = async_to_streamed_response_wrapper(
            external_price_id.fetch,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Union, Iterable, Optional
from datetime import datetime
from typing_extensions import Literal, Required, Annotated, TypeAlias, TypedDict

from .._types import SequenceNotStr
from .._utils import PropertyInfo
from .shared_params.new_floating_bulk_price import NewFloatingBulkPrice
from .shared_params.new_floating_unit_price import NewFloatingUnitPrice
from .shared_params.new_floating_matrix_price import NewFloatingMatrixPrice
from .shared_params.new_floating_tiered_price import NewFloatingTieredPrice
from .shared_params.new_floating_package_price import NewFloatingPackagePrice
from .shared_params.unit_conversion_rate_config import UnitConversionRateConfig
from .shared_params.tiered_conversion_rate_config import TieredConversionRateConfig
from .shared_params.new_billing_cycle_configuration import NewBillingCycleConfiguration
from .shared_params.new_floating_grouped_tiered_price import NewFloatingGroupedTieredPrice
from .shared_params.new_floating_tiered_package_price import NewFloatingTieredPackagePrice
from .shared_params.new_dimensional_price_configuration import NewDimensionalPriceConfiguration
from .shared_params.new_floating_minimum_composite_price import NewFloatingMinimumCompositePrice
from .shared_params.new_floating_unit_with_percent_price import NewFloatingUnitWithPercentPrice
from .shared_params.new_floating_grouped_allocation_price import NewFloatingGroupedAllocationPrice
from .shared_params.new_floating_bulk_with_proration_price import NewFloatingBulkWithProrationPrice
from .shared_params.new_floating_tiered_with_minimum_price import NewFloatingTieredWithMinimumPrice
from .shared_params.new_floating_unit_with_proration_price import NewFloatingUnitWithProrationPrice
from .shared_params.new_floating_tiered_with_proration_price import NewFloatingTieredWithProrationPrice
from .shared_params.new_floating_grouped_tiered_package_price import NewFloatingGroupedTieredPackagePrice
from .shared_params.new_floating_matrix_with_allocation_price import NewFloatingMatrixWithAllocationPrice
from .shared_params.new_floating_threshold_total_amount_price import NewFloatingThresholdTotalAmountPrice
from .shared_params.new_floating_cumulative_grouped_bulk_price import NewFloatingCumulativeGroupedBulkPrice
from .shared_params.new_floating_package_with_allocation_price import NewFloatingPackageWithAllocationPrice
from .shared_params.new_floating_matrix_with_display_name_price import NewFloatingMatrixWithDisplayNamePrice
from .shared_params.new_floating_max_group_tiered_package_price import NewFloatingMaxGroupTieredPackagePrice
from .shared_params.new_floating_tiered_package_with_minimum_price import NewFloatingTieredPackageWithMinimumPrice
from .shared_params.new_floating_grouped_with_metered_minimum_price import NewFloatingGroupedWithMeteredMinimumPrice
from .shared_params.new_floating_grouped_with_prorated_minimum_price import NewFloatingGroupedWithProratedMinimumPrice
from .shared_params.new_floating_scalable_matrix_with_unit_pricing_price import (
    NewFloatingScalableMatrixWithUnitPricingPrice,
)
from .shared_params.new_floating_scalable_matrix_with_tiered_pricing_price import (
    NewFloatingScalableMatrixWithTieredPricingPrice,
)

__all__ = [
    "PriceEvaluateMultipleParams",
    "PriceEvaluation",
    "PriceEvaluationPrice",
    "PriceEvaluationPriceNewFloatingGroupedWithMinMaxThresholdsPrice",
    "PriceEvaluationPriceNewFloatingGroupedWithMinMaxThresholdsPriceGroupedWithMinMaxThresholdsConfig",
    "PriceEvaluationPriceNewFloatingGroupedWithMinMaxThresholdsPriceConversionRateConfig",
]


class PriceEvaluateMultipleParams(TypedDict, total=False):
    timeframe_end: Required[Annotated[Union[str, datetime], PropertyInfo(format="iso8601")]]
    """The exclusive upper bound for event timestamps"""

    timeframe_start: Required[Annotated[Union[str, datetime], PropertyInfo(format="iso8601")]]
    """The inclusive lower bound for event timestamps"""

    customer_id: Optional[str]
    """The ID of the customer to which this evaluation is scoped."""

    external_customer_id: Optional[str]
    """The external customer ID of the customer to which this evaluation is scoped."""

    price_evaluations: Iterable[PriceEvaluation]
    """List of prices to evaluate (max 100)"""


class PriceEvaluationPriceNewFloatingGroupedWithMinMaxThresholdsPriceGroupedWithMinMaxThresholdsConfig(
    TypedDict, total=False
):
    grouping_key: Required[str]
    """The event property used to group before applying thresholds"""

    maximum_charge: Required[str]
    """The maximum amount to charge each group"""

    minimum_charge: Required[str]
    """The minimum amount to charge each group, regardless of usage"""

    per_unit_rate: Required[str]
    """The base price charged per group"""


PriceEvaluationPriceNewFloatingGroupedWithMinMaxThresholdsPriceConversionRateConfig: TypeAlias = Union[
    UnitConversionRateConfig, TieredConversionRateConfig
]


class PriceEvaluationPriceNewFloatingGroupedWithMinMaxThresholdsPrice(TypedDict, total=False):
    cadence: Required[Literal["annual", "semi_annual", "monthly", "quarterly", "one_time", "custom"]]
    """The cadence to bill for this price on."""

    currency: Required[str]
    """An ISO 4217 currency string for which this price is billed in."""

    grouped_with_min_max_thresholds_config: Required[
        PriceEvaluationPriceNewFloatingGroupedWithMinMaxThresholdsPriceGroupedWithMinMaxThresholdsConfig
    ]
    """Configuration for grouped_with_min_max_thresholds pricing"""

    item_id: Required[str]
    """The id of the item the price will be associated with."""

    model_type: Required[Literal["grouped_with_min_max_thresholds"]]
    """The pricing model type"""

    name: Required[str]
    """The name of the price."""

    billable_metric_id: Optional[str]
    """The id of the billable metric for the price.

    Only needed if the price is usage-based.
    """

    billed_in_advance: Optional[bool]
    """
    If the Price represents a fixed cost, the price will be billed in-advance if
    this is true, and in-arrears if this is false.
    """

    billing_cycle_configuration: Optional[NewBillingCycleConfiguration]
    """
    For custom cadence: specifies the duration of the billing period in days or
    months.
    """

    conversion_rate: Optional[float]
    """The per unit conversion rate of the price currency to the invoicing currency."""

    conversion_rate_config: Optional[
        PriceEvaluationPriceNewFloatingGroupedWithMinMaxThresholdsPriceConversionRateConfig
    ]
    """The configuration for the rate of the price currency to the invoicing currency."""

    dimensional_price_configuration: Optional[NewDimensionalPriceConfiguration]
    """For dimensional price: specifies a price group and dimension values"""

    external_price_id: Optional[str]
    """An alias for the price."""

    fixed_price_quantity: Optional[float]
    """
    If the Price represents a fixed cost, this represents the quantity of units
    applied.
    """

    invoice_grouping_key: Optional[str]
    """The property used to group this price on an invoice"""

    invoicing_cycle_configuration: Optional[NewBillingCycleConfiguration]
    """Within each billing cycle, specifies the cadence at which invoices are produced.

    If unspecified, a single invoice is produced per billing cycle.
    """

    metadata: Optional[Dict[str, Optional[str]]]
    """User-specified key/value pairs for the resource.

    Individual keys can be removed by setting the value to `null`, and the entire
    metadata mapping can be cleared by setting `metadata` to `null`.
    """


PriceEvaluationPrice: TypeAlias = Union[
    NewFloatingUnitPrice,
    NewFloatingTieredPrice,
    NewFloatingBulkPrice,
    NewFloatingPackagePrice,
    NewFloatingMatrixPrice,
    NewFloatingThresholdTotalAmountPrice,
    NewFloatingTieredPackagePrice,
    NewFloatingTieredWithMinimumPrice,
    NewFloatingGroupedTieredPrice,
    NewFloatingTieredPackageWithMinimumPrice,
    NewFloatingPackageWithAllocationPrice,
    NewFloatingUnitWithPercentPrice,
    NewFloatingMatrixWithAllocationPrice,
    NewFloatingTieredWithProrationPrice,
    NewFloatingUnitWithProrationPrice,
    NewFloatingGroupedAllocationPrice,
    NewFloatingBulkWithProrationPrice,
    NewFloatingGroupedWithProratedMinimumPrice,
    NewFloatingGroupedWithMeteredMinimumPrice,
    PriceEvaluationPriceNewFloatingGroupedWithMinMaxThresholdsPrice,
    NewFloatingMatrixWithDisplayNamePrice,
    NewFloatingGroupedTieredPackagePrice,
    NewFloatingMaxGroupTieredPackagePrice,
    NewFloatingScalableMatrixWithUnitPricingPrice,
    NewFloatingScalableMatrixWithTieredPricingPrice,
    NewFloatingCumulativeGroupedBulkPrice,
    NewFloatingMinimumCompositePrice,
]


class PriceEvaluation(TypedDict, total=False):
    external_price_id: Optional[str]
    """The external ID of a price to evaluate that exists in your Orb account."""

    filter: Optional[str]
    """
    A boolean
    [computed property](/extensibility/advanced-metrics#computed-properties) used to
    filter the underlying billable metric
    """

    grouping_keys: SequenceNotStr[str]
    """
    Properties (or
    [computed properties](/extensibility/advanced-metrics#computed-properties)) used
    to group the underlying billable metric
    """

    price: Optional[PriceEvaluationPrice]
    """New floating price request body params."""

    price_id: Optional[str]
    """The ID of a price to evaluate that exists in your Orb account."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List

from ..._models import BaseModel

__all__ = ["DimensionalPriceConfiguration"]


class DimensionalPriceConfiguration(BaseModel):
    dimension_values: List[str]

    dimensional_price_group_id: str



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Optional
from typing_extensions import Required, TypedDict

from ..._types import SequenceNotStr

__all__ = ["NewDimensionalPriceConfiguration"]


class NewDimensionalPriceConfiguration(TypedDict, total=False):
    dimension_values: Required[SequenceNotStr[str]]
    """
    The list of dimension values matching (in order) the dimensions of the price
    group
    """

    dimensional_price_group_id: Optional[str]
    """The id of the dimensional price group to include this price in"""

    external_dimensional_price_group_id: Optional[str]
    """The external id of the dimensional price group to include this price in"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal

from ..._models import BaseModel
from .conversion_rate_tiered_config import ConversionRateTieredConfig

__all__ = ["TieredConversionRateConfig"]


class TieredConversionRateConfig(BaseModel):
    conversion_rate_type: Literal["tiered"]

    tiered_config: ConversionRateTieredConfig



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union, Iterable, Optional
from datetime import date, datetime
from typing_extensions import Literal

import httpx

from .. import _legacy_response
from ..types import credit_note_list_params, credit_note_create_params
from .._types import Body, Omit, Query, Headers, NotGiven, omit, not_given
from .._utils import maybe_transform, async_maybe_transform
from .._compat import cached_property
from .._resource import SyncAPIResource, AsyncAPIResource
from .._response import to_streamed_response_wrapper, async_to_streamed_response_wrapper
from ..pagination import SyncPage, AsyncPage
from .._base_client import AsyncPaginator, make_request_options
from ..types.shared.credit_note import CreditNote

__all__ = ["CreditNotes", "AsyncCreditNotes"]


class CreditNotes(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> CreditNotesWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/orbcorp/orb-python#accessing-raw-response-data-eg-headers
        """
        return CreditNotesWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> CreditNotesWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/orbcorp/orb-python#with_streaming_response
        """
        return CreditNotesWithStreamingResponse(self)

    def create(
        self,
        *,
        line_items: Iterable[credit_note_create_params.LineItem],
        reason: Literal["duplicate", "fraudulent", "order_change", "product_unsatisfactory"],
        end_date: Union[str, date, None] | Omit = omit,
        memo: Optional[str] | Omit = omit,
        start_date: Union[str, date, None] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> CreditNote:
        """
        This endpoint is used to create a single
        [`Credit Note`](/invoicing/credit-notes).

        The credit note service period configuration supports two explicit modes:

        1. Global service periods: Specify start_date and end_date at the credit note
           level. These dates will be applied to all line items uniformly.

        2. Individual service periods: Specify start_date and end_date for each line
           item. When using this mode, ALL line items must have individual periods
           specified.

        3. Default behavior: If no service periods are specified (neither global nor
           individual), the original invoice line item service periods will be used.

        Note: Mixing global and individual service periods in the same request is not
        allowed to prevent confusion.

        Service period dates are normalized to the start of the day in the customer's
        timezone to ensure consistent handling across different timezones.

        Date Format: Use start_date and end_date with format "YYYY-MM-DD" (e.g.,
        "2023-09-22") to match other Orb APIs like /v1/invoice_line_items.

        Note: Both start_date and end_date are inclusive - the service period will cover
        both the start date and end date completely (from start of start_date to end of
        end_date).

        Args:
          reason: An optional reason for the credit note.

          end_date: A date string to specify the global credit note service period end date in the
              customer's timezone. This will be applied to all line items that don't have
              their own individual service periods specified. If not provided, line items will
              use their original invoice line item service periods. This date is inclusive.

          memo: An optional memo to attach to the credit note.

          start_date: A date string to specify the global credit note service period start date in the
              customer's timezone. This will be applied to all line items that don't have
              their own individual service periods specified. If not provided, line items will
              use their original invoice line item service periods. This date is inclusive.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return self._post(
            "/credit_notes",
            body=maybe_transform(
                {
                    "line_items": line_items,
                    "reason": reason,
                    "end_date": end_date,
                    "memo": memo,
                    "start_date": start_date,
                },
                credit_note_create_params.CreditNoteCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=CreditNote,
        )

    def list(
        self,
        *,
        created_at_gt: Union[str, datetime, None] | Omit = omit,
        created_at_gte: Union[str, datetime, None] | Omit = omit,
        created_at_lt: Union[str, datetime, None] | Omit = omit,
        created_at_lte: Union[str, datetime, None] | Omit = omit,
        cursor: Optional[str] | Omit = omit,
        limit: int | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SyncPage[CreditNote]:
        """Get a paginated list of CreditNotes.

        Users can also filter by customer_id,
        subscription_id, or external_customer_id. The credit notes will be returned in
        reverse chronological order by `creation_time`.

        Args:
          cursor: Cursor for pagination. This can be populated by the `next_cursor` value returned
              from the initial request.

          limit: The number of items to fetch. Defaults to 20.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get_api_list(
            "/credit_notes",
            page=SyncPage[CreditNote],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "created_at_gt": created_at_gt,
                        "created_at_gte": created_at_gte,
                        "created_at_lt": created_at_lt,
                        "created_at_lte": created_at_lte,
                        "cursor": cursor,
                        "limit": limit,
                    },
                    credit_note_list_params.CreditNoteListParams,
                ),
            ),
            model=CreditNote,
        )

    def fetch(
        self,
        credit_note_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> CreditNote:
        """
        This endpoint is used to fetch a single [`Credit Note`](/invoicing/credit-notes)
        given an identifier.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not credit_note_id:
            raise ValueError(f"Expected a non-empty value for `credit_note_id` but received {credit_note_id!r}")
        return self._get(
            f"/credit_notes/{credit_note_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=CreditNote,
        )


class AsyncCreditNotes(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncCreditNotesWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/orbcorp/orb-python#accessing-raw-response-data-eg-headers
        """
        return AsyncCreditNotesWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncCreditNotesWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/orbcorp/orb-python#with_streaming_response
        """
        return AsyncCreditNotesWithStreamingResponse(self)

    async def create(
        self,
        *,
        line_items: Iterable[credit_note_create_params.LineItem],
        reason: Literal["duplicate", "fraudulent", "order_change", "product_unsatisfactory"],
        end_date: Union[str, date, None] | Omit = omit,
        memo: Optional[str] | Omit = omit,
        start_date: Union[str, date, None] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> CreditNote:
        """
        This endpoint is used to create a single
        [`Credit Note`](/invoicing/credit-notes).

        The credit note service period configuration supports two explicit modes:

        1. Global service periods: Specify start_date and end_date at the credit note
           level. These dates will be applied to all line items uniformly.

        2. Individual service periods: Specify start_date and end_date for each line
           item. When using this mode, ALL line items must have individual periods
           specified.

        3. Default behavior: If no service periods are specified (neither global nor
           individual), the original invoice line item service periods will be used.

        Note: Mixing global and individual service periods in the same request is not
        allowed to prevent confusion.

        Service period dates are normalized to the start of the day in the customer's
        timezone to ensure consistent handling across different timezones.

        Date Format: Use start_date and end_date with format "YYYY-MM-DD" (e.g.,
        "2023-09-22") to match other Orb APIs like /v1/invoice_line_items.

        Note: Both start_date and end_date are inclusive - the service period will cover
        both the start date and end date completely (from start of start_date to end of
        end_date).

        Args:
          reason: An optional reason for the credit note.

          end_date: A date string to specify the global credit note service period end date in the
              customer's timezone. This will be applied to all line items that don't have
              their own individual service periods specified. If not provided, line items will
              use their original invoice line item service periods. This date is inclusive.

          memo: An optional memo to attach to the credit note.

          start_date: A date string to specify the global credit note service period start date in the
              customer's timezone. This will be applied to all line items that don't have
              their own individual service periods specified. If not provided, line items will
              use their original invoice line item service periods. This date is inclusive.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return await self._post(
            "/credit_notes",
            body=await async_maybe_transform(
                {
                    "line_items": line_items,
                    "reason": reason,
                    "end_date": end_date,
                    "memo": memo,
                    "start_date": start_date,
                },
                credit_note_create_params.CreditNoteCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=CreditNote,
        )

    def list(
        self,
        *,
        created_at_gt: Union[str, datetime, None] | Omit = omit,
        created_at_gte: Union[str, datetime, None] | Omit = omit,
        created_at_lt: Union[str, datetime, None] | Omit = omit,
        created_at_lte: Union[str, datetime, None] | Omit = omit,
        cursor: Optional[str] | Omit = omit,
        limit: int | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> AsyncPaginator[CreditNote, AsyncPage[CreditNote]]:
        """Get a paginated list of CreditNotes.

        Users can also filter by customer_id,
        subscription_id, or external_customer_id. The credit notes will be returned in
        reverse chronological order by `creation_time`.

        Args:
          cursor: Cursor for pagination. This can be populated by the `next_cursor` value returned
              from the initial request.

          limit: The number of items to fetch. Defaults to 20.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get_api_list(
            "/credit_notes",
            page=AsyncPage[CreditNote],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "created_at_gt": created_at_gt,
                        "created_at_gte": created_at_gte,
                        "created_at_lt": created_at_lt,
                        "created_at_lte": created_at_lte,
                        "cursor": cursor,
                        "limit": limit,
                    },
                    credit_note_list_params.CreditNoteListParams,
                ),
            ),
            model=CreditNote,
        )

    async def fetch(
        self,
        credit_note_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> CreditNote:
        """
        This endpoint is used to fetch a single [`Credit Note`](/invoicing/credit-notes)
        given an identifier.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not credit_note_id:
            raise ValueError(f"Expected a non-empty value for `credit_note_id` but received {credit_note_id!r}")
        return await self._get(
            f"/credit_notes/{credit_note_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=CreditNote,
        )


class CreditNotesWithRawResponse:
    def __init__(self, credit_notes: CreditNotes) -> None:
        self._credit_notes = credit_notes

        self.create = _legacy_response.to_raw_response_wrapper(
            credit_notes.create,
        )
        self.list = _legacy_response.to_raw_response_wrapper(
            credit_notes.list,
        )
        self.fetch = _legacy_response.to_raw_response_wrapper(
            credit_notes.fetch,
        )


class AsyncCreditNotesWithRawResponse:
    def __init__(self, credit_notes: AsyncCreditNotes) -> None:
        self._credit_notes = credit_notes

        self.create = _legacy_response.async_to_raw_response_wrapper(
            credit_notes.create,
        )
        self.list = _legacy_response.async_to_raw_response_wrapper(
            credit_notes.list,
        )
        self.fetch = _legacy_response.async_to_raw_response_wrapper(
            credit_notes.fetch,
        )


class CreditNotesWithStreamingResponse:
    def __init__(self, credit_notes: CreditNotes) -> None:
        self._credit_notes = credit_notes

        self.create = to_streamed_response_wrapper(
            credit_notes.create,
        )
        self.list = to_streamed_response_wrapper(
            credit_notes.list,
        )
        self.fetch = to_streamed_response_wrapper(
            credit_notes.fetch,
        )


class AsyncCreditNotesWithStreamingResponse:
    def __init__(self, credit_notes: AsyncCreditNotes) -> None:
        self._credit_notes = credit_notes

        self.create = async_to_streamed_response_wrapper(
            credit_notes.create,
        )
        self.list = async_to_streamed_response_wrapper(
            credit_notes.list,
        )
        self.fetch = async_to_streamed_response_wrapper(
            credit_notes.fetch,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Required, TypedDict

__all__ = ["ConversionRateUnitConfig"]


class ConversionRateUnitConfig(TypedDict, total=False):
    unit_amount: Required[str]
    """Amount per unit of overage"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional

from ..._models import BaseModel

__all__ = ["TaxAmount"]


class TaxAmount(BaseModel):
    amount: str
    """The amount of additional tax incurred by this tax rate."""

    tax_rate_description: str
    """The human-readable description of the applied tax rate."""

    tax_rate_percentage: Optional[str] = None
    """The tax rate percentage, out of 100."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Optional

import httpx

from .. import _legacy_response
from ..types import subscription_change_apply_params
from .._types import Body, Omit, Query, Headers, NotGiven, omit, not_given
from .._utils import maybe_transform, async_maybe_transform
from .._compat import cached_property
from .._resource import SyncAPIResource, AsyncAPIResource
from .._response import to_streamed_response_wrapper, async_to_streamed_response_wrapper
from .._base_client import make_request_options
from ..types.subscription_change_apply_response import SubscriptionChangeApplyResponse
from ..types.subscription_change_cancel_response import SubscriptionChangeCancelResponse
from ..types.subscription_change_retrieve_response import SubscriptionChangeRetrieveResponse

__all__ = ["SubscriptionChanges", "AsyncSubscriptionChanges"]


class SubscriptionChanges(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> SubscriptionChangesWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/orbcorp/orb-python#accessing-raw-response-data-eg-headers
        """
        return SubscriptionChangesWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> SubscriptionChangesWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/orbcorp/orb-python#with_streaming_response
        """
        return SubscriptionChangesWithStreamingResponse(self)

    def retrieve(
        self,
        subscription_change_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SubscriptionChangeRetrieveResponse:
        """
        This endpoint returns a subscription change given an identifier.

        A subscription change is created by including
        `Create-Pending-Subscription-Change: True` in the header of a subscription
        mutation API call (e.g.
        [create subscription endpoint](/api-reference/subscription/create-subscription),
        [schedule plan change endpoint](/api-reference/subscription/schedule-plan-change),
        ...). The subscription change will be referenced by the
        `pending_subscription_change` field in the response.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not subscription_change_id:
            raise ValueError(
                f"Expected a non-empty value for `subscription_change_id` but received {subscription_change_id!r}"
            )
        return self._get(
            f"/subscription_changes/{subscription_change_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=SubscriptionChangeRetrieveResponse,
        )

    def apply(
        self,
        subscription_change_id: str,
        *,
        description: Optional[str] | Omit = omit,
        previously_collected_amount: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> SubscriptionChangeApplyResponse:
        """Apply a subscription change to perform the intended action.

        If a positive amount
        is passed with a request to this endpoint, any eligible invoices that were
        created will be issued immediately if they only contain in-advance fees.

        Args:
          description: Description to apply to the balance transaction representing this credit.

          previously_collected_amount: Amount already collected to apply to the customer's balance.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        if not subscription_change_id:
            raise ValueError(
                f"Expected a non-empty value for `subscription_change_id` but received {subscription_change_id!r}"
            )
        return self._post(
            f"/subscription_changes/{subscription_change_id}/apply",
            body=maybe_transform(
                {
                    "description": description,
                    "previously_collected_amount": previously_collected_amount,
                },
                subscription_change_apply_params.SubscriptionChangeApplyParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=SubscriptionChangeApplyResponse,
        )

    def cancel(
        self,
        subscription_change_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> SubscriptionChangeCancelResponse:
        """Cancel a subscription change.

        The change can no longer be applied. A
        subscription can only have one "pending" change at a time - use this endpoint to
        cancel an existing change before creating a new one.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        if not subscription_change_id:
            raise ValueError(
                f"Expected a non-empty value for `subscription_change_id` but received {subscription_change_id!r}"
            )
        return self._post(
            f"/subscription_changes/{subscription_change_id}/cancel",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=SubscriptionChangeCancelResponse,
        )


class AsyncSubscriptionChanges(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncSubscriptionChangesWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/orbcorp/orb-python#accessing-raw-response-data-eg-headers
        """
        return AsyncSubscriptionChangesWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncSubscriptionChangesWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/orbcorp/orb-python#with_streaming_response
        """
        return AsyncSubscriptionChangesWithStreamingResponse(self)

    async def retrieve(
        self,
        subscription_change_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SubscriptionChangeRetrieveResponse:
        """
        This endpoint returns a subscription change given an identifier.

        A subscription change is created by including
        `Create-Pending-Subscription-Change: True` in the header of a subscription
        mutation API call (e.g.
        [create subscription endpoint](/api-reference/subscription/create-subscription),
        [schedule plan change endpoint](/api-reference/subscription/schedule-plan-change),
        ...). The subscription change will be referenced by the
        `pending_subscription_change` field in the response.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not subscription_change_id:
            raise ValueError(
                f"Expected a non-empty value for `subscription_change_id` but received {subscription_change_id!r}"
            )
        return await self._get(
            f"/subscription_changes/{subscription_change_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=SubscriptionChangeRetrieveResponse,
        )

    async def apply(
        self,
        subscription_change_id: str,
        *,
        description: Optional[str] | Omit = omit,
        previously_collected_amount: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> SubscriptionChangeApplyResponse:
        """Apply a subscription change to perform the intended action.

        If a positive amount
        is passed with a request to this endpoint, any eligible invoices that were
        created will be issued immediately if they only contain in-advance fees.

        Args:
          description: Description to apply to the balance transaction representing this credit.

          previously_collected_amount: Amount already collected to apply to the customer's balance.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        if not subscription_change_id:
            raise ValueError(
                f"Expected a non-empty value for `subscription_change_id` but received {subscription_change_id!r}"
            )
        return await self._post(
            f"/subscription_changes/{subscription_change_id}/apply",
            body=await async_maybe_transform(
                {
                    "description": description,
                    "previously_collected_amount": previously_collected_amount,
                },
                subscription_change_apply_params.SubscriptionChangeApplyParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=SubscriptionChangeApplyResponse,
        )

    async def cancel(
        self,
        subscription_change_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
        idempotency_key: str | None = None,
    ) -> SubscriptionChangeCancelResponse:
        """Cancel a subscription change.

        The change can no longer be applied. A
        subscription can only have one "pending" change at a time - use this endpoint to
        cancel an existing change before creating a new one.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        if not subscription_change_id:
            raise ValueError(
                f"Expected a non-empty value for `subscription_change_id` but received {subscription_change_id!r}"
            )
        return await self._post(
            f"/subscription_changes/{subscription_change_id}/cancel",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=SubscriptionChangeCancelResponse,
        )


class SubscriptionChangesWithRawResponse:
    def __init__(self, subscription_changes: SubscriptionChanges) -> None:
        self._subscription_changes = subscription_changes

        self.retrieve = _legacy_response.to_raw_response_wrapper(
            subscription_changes.retrieve,
        )
        self.apply = _legacy_response.to_raw_response_wrapper(
            subscription_changes.apply,
        )
        self.cancel = _legacy_response.to_raw_response_wrapper(
            subscription_changes.cancel,
        )


class AsyncSubscriptionChangesWithRawResponse:
    def __init__(self, subscription_changes: AsyncSubscriptionChanges) -> None:
        self._subscription_changes = subscription_changes

        self.retrieve = _legacy_response.async_to_raw_response_wrapper(
            subscription_changes.retrieve,
        )
        self.apply = _legacy_response.async_to_raw_response_wrapper(
            subscription_changes.apply,
        )
        self.cancel = _legacy_response.async_to_raw_response_wrapper(
            subscription_changes.cancel,
        )


class SubscriptionChangesWithStreamingResponse:
    def __init__(self, subscription_changes: SubscriptionChanges) -> None:
        self._subscription_changes = subscription_changes

        self.retrieve = to_streamed_response_wrapper(
            subscription_changes.retrieve,
        )
        self.apply = to_streamed_response_wrapper(
            subscription_changes.apply,
        )
        self.cancel = to_streamed_response_wrapper(
            subscription_changes.cancel,
        )


class AsyncSubscriptionChangesWithStreamingResponse:
    def __init__(self, subscription_changes: AsyncSubscriptionChanges) -> None:
        self._subscription_changes = subscription_changes

        self.retrieve = async_to_streamed_response_wrapper(
            subscription_changes.retrieve,
        )
        self.apply = async_to_streamed_response_wrapper(
            subscription_changes.apply,
        )
        self.cancel = async_to_streamed_response_wrapper(
            subscription_changes.cancel,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from .item import Item as Item
from .plan import Plan as Plan
from .alert import Alert as Alert
from .coupon import Coupon as Coupon
from .shared import (
    Tier as Tier,
    Price as Price,
    Address as Address,
    Invoice as Invoice,
    Maximum as Maximum,
    Minimum as Minimum,
    BulkTier as BulkTier,
    Discount as Discount,
    ItemSlim as ItemSlim,
    TaxAmount as TaxAmount,
    Allocation as Allocation,
    BulkConfig as BulkConfig,
    CreditNote as CreditNote,
    NewMaximum as NewMaximum,
    NewMinimum as NewMinimum,
    UnitConfig as UnitConfig,
    InvoiceTiny as InvoiceTiny,
    MatrixValue as MatrixValue,
    MatrixConfig as MatrixConfig,
    PerPriceCost as PerPriceCost,
    TieredConfig as TieredConfig,
    CustomerTaxID as CustomerTaxID,
    PackageConfig as PackageConfig,
    PriceInterval as PriceInterval,
    TrialDiscount as TrialDiscount,
    UsageDiscount as UsageDiscount,
    AggregatedCost as AggregatedCost,
    AmountDiscount as AmountDiscount,
    CreditNoteTiny as CreditNoteTiny,
    MaximumInterval as MaximumInterval,
    MinimumInterval as MinimumInterval,
    TierSubLineItem as TierSubLineItem,
    CouponRedemption as CouponRedemption,
    CustomerMinified as CustomerMinified,
    CustomExpiration as CustomExpiration,
    NewPlanBulkPrice as NewPlanBulkPrice,
    NewPlanUnitPrice as NewPlanUnitPrice,
    NewUsageDiscount as NewUsageDiscount,
    OtherSubLineItem as OtherSubLineItem,
    MatrixSubLineItem as MatrixSubLineItem,
    NewAmountDiscount as NewAmountDiscount,
    AdjustmentInterval as AdjustmentInterval,
    BillableMetricTiny as BillableMetricTiny,
    ConversionRateTier as ConversionRateTier,
    NewAllocationPrice as NewAllocationPrice,
    NewPlanMatrixPrice as NewPlanMatrixPrice,
    NewPlanTieredPrice as NewPlanTieredPrice,
    PaginationMetadata as PaginationMetadata,
    PercentageDiscount as PercentageDiscount,
    NewPlanPackagePrice as NewPlanPackagePrice,
    SubLineItemGrouping as SubLineItemGrouping,
    InvoiceLevelDiscount as InvoiceLevelDiscount,
    NewFloatingBulkPrice as NewFloatingBulkPrice,
    NewFloatingUnitPrice as NewFloatingUnitPrice,
    SubscriptionMinified as SubscriptionMinified,
    TransformPriceFilter as TransformPriceFilter,
    NewPercentageDiscount as NewPercentageDiscount,
    SubscriptionTrialInfo as SubscriptionTrialInfo,
    UsageDiscountInterval as UsageDiscountInterval,
    AmountDiscountInterval as AmountDiscountInterval,
    NewFloatingMatrixPrice as NewFloatingMatrixPrice,
    NewFloatingTieredPrice as NewFloatingTieredPrice,
    NewFloatingPackagePrice as NewFloatingPackagePrice,
    SubLineItemMatrixConfig as SubLineItemMatrixConfig,
    BillingCycleRelativeDate as BillingCycleRelativeDate,
    ConversionRateUnitConfig as ConversionRateUnitConfig,
    UnitConversionRateConfig as UnitConversionRateConfig,
    BillingCycleConfiguration as BillingCycleConfiguration,
    MonetaryMaximumAdjustment as MonetaryMaximumAdjustment,
    MonetaryMinimumAdjustment as MonetaryMinimumAdjustment,
    NewPlanGroupedTieredPrice as NewPlanGroupedTieredPrice,
    NewPlanTieredPackagePrice as NewPlanTieredPackagePrice,
    ConversionRateTieredConfig as ConversionRateTieredConfig,
    FixedFeeQuantityTransition as FixedFeeQuantityTransition,
    MatrixWithAllocationConfig as MatrixWithAllocationConfig,
    PercentageDiscountInterval as PercentageDiscountInterval,
    PlanPhaseMaximumAdjustment as PlanPhaseMaximumAdjustment,
    PlanPhaseMinimumAdjustment as PlanPhaseMinimumAdjustment,
    SubscriptionChangeMinified as SubscriptionChangeMinified,
    TieredConversionRateConfig as TieredConversionRateConfig,
    NewPlanUnitWithPercentPrice as NewPlanUnitWithPercentPrice,
    ChangedSubscriptionResources as ChangedSubscriptionResources,
    NewBillingCycleConfiguration as NewBillingCycleConfiguration,
    NewPlanMinimumCompositePrice as NewPlanMinimumCompositePrice,
    DimensionalPriceConfiguration as DimensionalPriceConfiguration,
    FixedFeeQuantityScheduleEntry as FixedFeeQuantityScheduleEntry,
    NewFloatingGroupedTieredPrice as NewFloatingGroupedTieredPrice,
    NewFloatingTieredPackagePrice as NewFloatingTieredPackagePrice,
    NewPlanBulkWithProrationPrice as NewPlanBulkWithProrationPrice,
    NewPlanGroupedAllocationPrice as NewPlanGroupedAllocationPrice,
    NewPlanTieredWithMinimumPrice as NewPlanTieredWithMinimumPrice,
    NewPlanUnitWithProrationPrice as NewPlanUnitWithProrationPrice,
    BillingCycleAnchorConfiguration as BillingCycleAnchorConfiguration,
    MonetaryUsageDiscountAdjustment as MonetaryUsageDiscountAdjustment,
    NewFloatingUnitWithPercentPrice as NewFloatingUnitWithPercentPrice,
    MonetaryAmountDiscountAdjustment as MonetaryAmountDiscountAdjustment,
    NewDimensionalPriceConfiguration as NewDimensionalPriceConfiguration,
    NewFloatingMinimumCompositePrice as NewFloatingMinimumCompositePrice,
    NewPlanGroupedTieredPackagePrice as NewPlanGroupedTieredPackagePrice,
    NewPlanMatrixWithAllocationPrice as NewPlanMatrixWithAllocationPrice,
    NewPlanThresholdTotalAmountPrice as NewPlanThresholdTotalAmountPrice,
    PlanPhaseUsageDiscountAdjustment as PlanPhaseUsageDiscountAdjustment,
    NewFloatingBulkWithProrationPrice as NewFloatingBulkWithProrationPrice,
    NewFloatingGroupedAllocationPrice as NewFloatingGroupedAllocationPrice,
    NewFloatingTieredWithMinimumPrice as NewFloatingTieredWithMinimumPrice,
    NewFloatingUnitWithProrationPrice as NewFloatingUnitWithProrationPrice,
    NewPlanCumulativeGroupedBulkPrice as NewPlanCumulativeGroupedBulkPrice,
    NewPlanMatrixWithDisplayNamePrice as NewPlanMatrixWithDisplayNamePrice,
    NewPlanMaxGroupTieredPackagePrice as NewPlanMaxGroupTieredPackagePrice,
    NewPlanPackageWithAllocationPrice as NewPlanPackageWithAllocationPrice,
    PlanPhaseAmountDiscountAdjustment as PlanPhaseAmountDiscountAdjustment,
    NewFloatingTieredWithProrationPrice as NewFloatingTieredWithProrationPrice,
    MonetaryPercentageDiscountAdjustment as MonetaryPercentageDiscountAdjustment,
    NewFloatingGroupedTieredPackagePrice as NewFloatingGroupedTieredPackagePrice,
    NewFloatingMatrixWithAllocationPrice as NewFloatingMatrixWithAllocationPrice,
    NewFloatingThresholdTotalAmountPrice as NewFloatingThresholdTotalAmountPrice,
    NewPlanTieredPackageWithMinimumPrice as NewPlanTieredPackageWithMinimumPrice,
    NewFloatingCumulativeGroupedBulkPrice as NewFloatingCumulativeGroupedBulkPrice,
    NewFloatingMatrixWithDisplayNamePrice as NewFloatingMatrixWithDisplayNamePrice,
    NewFloatingMaxGroupTieredPackagePrice as NewFloatingMaxGroupTieredPackagePrice,
    NewFloatingPackageWithAllocationPrice as NewFloatingPackageWithAllocationPrice,
    NewPlanGroupedWithMeteredMinimumPrice as NewPlanGroupedWithMeteredMinimumPrice,
    PlanPhasePercentageDiscountAdjustment as PlanPhasePercentageDiscountAdjustment,
    NewPlanGroupedWithProratedMinimumPrice as NewPlanGroupedWithProratedMinimumPrice,
    NewFloatingTieredPackageWithMinimumPrice as NewFloatingTieredPackageWithMinimumPrice,
    NewFloatingGroupedWithMeteredMinimumPrice as NewFloatingGroupedWithMeteredMinimumPrice,
    NewPlanScalableMatrixWithUnitPricingPrice as NewPlanScalableMatrixWithUnitPricingPrice,
    NewFloatingGroupedWithProratedMinimumPrice as NewFloatingGroupedWithProratedMinimumPrice,
    NewPlanScalableMatrixWithTieredPricingPrice as NewPlanScalableMatrixWithTieredPricingPrice,
    NewFloatingScalableMatrixWithUnitPricingPrice as NewFloatingScalableMatrixWithUnitPricingPrice,
    NewFloatingScalableMatrixWithTieredPricingPrice as NewFloatingScalableMatrixWithTieredPricingPrice,
)
from .customer import Customer as Customer
from .threshold import Threshold as Threshold
from .plan_version import PlanVersion as PlanVersion
from .subscription import Subscription as Subscription
from .subscriptions import Subscriptions as Subscriptions
from .billable_metric import BillableMetric as BillableMetric
from .threshold_param import ThresholdParam as ThresholdParam
from .item_list_params import ItemListParams as ItemListParams
from .plan_list_params import PlanListParams as PlanListParams
from .alert_list_params import AlertListParams as AlertListParams
from .price_list_params import PriceListParams as PriceListParams
from .coupon_list_params import CouponListParams as CouponListParams
from .item_create_params import ItemCreateParams as ItemCreateParams
from .item_update_params import ItemUpdateParams as ItemUpdateParams
from .metric_list_params import MetricListParams as MetricListParams
from .plan_create_params import PlanCreateParams as PlanCreateParams
from .plan_update_params import PlanUpdateParams as PlanUpdateParams
from .plan_version_phase import PlanVersionPhase as PlanVersionPhase
from .subscription_usage import SubscriptionUsage as SubscriptionUsage
from .address_input_param import AddressInputParam as AddressInputParam
from .alert_enable_params import AlertEnableParams as AlertEnableParams
from .alert_update_params import AlertUpdateParams as AlertUpdateParams
from .event_ingest_params import EventIngestParams as EventIngestParams
from .event_search_params import EventSearchParams as EventSearchParams
from .event_update_params import EventUpdateParams as EventUpdateParams
from .invoice_list_params import InvoiceListParams as InvoiceListParams
from .price_create_params import PriceCreateParams as PriceCreateParams
from .price_update_params import PriceUpdateParams as PriceUpdateParams
from .alert_disable_params import AlertDisableParams as AlertDisableParams
from .coupon_create_params import CouponCreateParams as CouponCreateParams
from .customer_list_params import CustomerListParams as CustomerListParams
from .evaluate_price_group import EvaluatePriceGroup as EvaluatePriceGroup
from .invoice_issue_params import InvoiceIssueParams as InvoiceIssueParams
from .metric_create_params import MetricCreateParams as MetricCreateParams
from .metric_update_params import MetricUpdateParams as MetricUpdateParams
from .mutated_subscription import MutatedSubscription as MutatedSubscription
from .event_ingest_response import EventIngestResponse as EventIngestResponse
from .event_search_response import EventSearchResponse as EventSearchResponse
from .event_update_response import EventUpdateResponse as EventUpdateResponse
from .invoice_create_params import InvoiceCreateParams as InvoiceCreateParams
from .invoice_update_params import InvoiceUpdateParams as InvoiceUpdateParams
from .price_evaluate_params import PriceEvaluateParams as PriceEvaluateParams
from .customer_create_params import CustomerCreateParams as CustomerCreateParams
from .customer_update_params import CustomerUpdateParams as CustomerUpdateParams
from .credit_note_list_params import CreditNoteListParams as CreditNoteListParams
from .dimensional_price_group import DimensionalPriceGroup as DimensionalPriceGroup
from .discount_override_param import DiscountOverrideParam as DiscountOverrideParam
from .price_evaluate_response import PriceEvaluateResponse as PriceEvaluateResponse
from .top_level_ping_response import TopLevelPingResponse as TopLevelPingResponse
from .dimensional_price_groups import DimensionalPriceGroups as DimensionalPriceGroups
from .event_deprecate_response import EventDeprecateResponse as EventDeprecateResponse
from .invoice_mark_paid_params import InvoiceMarkPaidParams as InvoiceMarkPaidParams
from .subscription_list_params import SubscriptionListParams as SubscriptionListParams
from .credit_note_create_params import CreditNoteCreateParams as CreditNoteCreateParams
from .subscription_cancel_params import SubscriptionCancelParams as SubscriptionCancelParams
from .subscription_create_params import SubscriptionCreateParams as SubscriptionCreateParams
from .subscription_update_params import SubscriptionUpdateParams as SubscriptionUpdateParams
from .invoice_fetch_upcoming_params import InvoiceFetchUpcomingParams as InvoiceFetchUpcomingParams
from .new_sphere_configuration_param import NewSphereConfigurationParam as NewSphereConfigurationParam
from .price_evaluate_multiple_params import PriceEvaluateMultipleParams as PriceEvaluateMultipleParams
from .beta_create_plan_version_params import BetaCreatePlanVersionParams as BetaCreatePlanVersionParams
from .customer_hierarchy_config_param import CustomerHierarchyConfigParam as CustomerHierarchyConfigParam
from .invoice_fetch_upcoming_response import InvoiceFetchUpcomingResponse as InvoiceFetchUpcomingResponse
from .invoice_line_item_create_params import InvoiceLineItemCreateParams as InvoiceLineItemCreateParams
from .new_tax_jar_configuration_param import NewTaxJarConfigurationParam as NewTaxJarConfigurationParam
from .subscription_fetch_costs_params import SubscriptionFetchCostsParams as SubscriptionFetchCostsParams
from .subscription_fetch_usage_params import SubscriptionFetchUsageParams as SubscriptionFetchUsageParams
from .accounting_provider_config_param import AccountingProviderConfigParam as AccountingProviderConfigParam
from .alert_create_for_customer_params import AlertCreateForCustomerParams as AlertCreateForCustomerParams
from .price_evaluate_multiple_response import PriceEvaluateMultipleResponse as PriceEvaluateMultipleResponse
from .subscription_change_apply_params import SubscriptionChangeApplyParams as SubscriptionChangeApplyParams
from .subscription_update_trial_params import SubscriptionUpdateTrialParams as SubscriptionUpdateTrialParams
from .invoice_line_item_create_response import InvoiceLineItemCreateResponse as InvoiceLineItemCreateResponse
from .new_reporting_configuration_param import NewReportingConfigurationParam as NewReportingConfigurationParam
from .new_subscription_bulk_price_param import NewSubscriptionBulkPriceParam as NewSubscriptionBulkPriceParam
from .new_subscription_unit_price_param import NewSubscriptionUnitPriceParam as NewSubscriptionUnitPriceParam
from .subscription_fetch_costs_response import SubscriptionFetchCostsResponse as SubscriptionFetchCostsResponse
from .subscription_redeem_coupon_params import SubscriptionRedeemCouponParams as SubscriptionRedeemCouponParams
from .subscription_trigger_phase_params import SubscriptionTriggerPhaseParams as SubscriptionTriggerPhaseParams
from .subscription_change_apply_response import SubscriptionChangeApplyResponse as SubscriptionChangeApplyResponse
from .subscription_fetch_schedule_params import SubscriptionFetchScheduleParams as SubscriptionFetchScheduleParams
from .dimensional_price_group_list_params import DimensionalPriceGroupListParams as DimensionalPriceGroupListParams
from .new_avalara_tax_configuration_param import NewAvalaraTaxConfigurationParam as NewAvalaraTaxConfigurationParam
from .new_subscription_matrix_price_param import NewSubscriptionMatrixPriceParam as NewSubscriptionMatrixPriceParam
from .new_subscription_tiered_price_param import NewSubscriptionTieredPriceParam as NewSubscriptionTieredPriceParam
from .subscription_change_cancel_response import SubscriptionChangeCancelResponse as SubscriptionChangeCancelResponse
from .subscription_price_intervals_params import SubscriptionPriceIntervalsParams as SubscriptionPriceIntervalsParams
from .alert_create_for_subscription_params import AlertCreateForSubscriptionParams as AlertCreateForSubscriptionParams
from .beta_set_default_plan_version_params import BetaSetDefaultPlanVersionParams as BetaSetDefaultPlanVersionParams
from .new_subscription_package_price_param import NewSubscriptionPackagePriceParam as NewSubscriptionPackagePriceParam
from .price_evaluate_preview_events_params import PriceEvaluatePreviewEventsParams as PriceEvaluatePreviewEventsParams
from .subscription_fetch_schedule_response import SubscriptionFetchScheduleResponse as SubscriptionFetchScheduleResponse
from .customer_update_by_external_id_params import CustomerUpdateByExternalIDParams as CustomerUpdateByExternalIDParams
from .dimensional_price_group_create_params import (
    DimensionalPriceGroupCreateParams as DimensionalPriceGroupCreateParams,
)
from .dimensional_price_group_update_params import (
    DimensionalPriceGroupUpdateParams as DimensionalPriceGroupUpdateParams,
)
from .subscription_change_retrieve_response import (
    SubscriptionChangeRetrieveResponse as SubscriptionChangeRetrieveResponse,
)
from .price_evaluate_preview_events_response import (
    PriceEvaluatePreviewEventsResponse as PriceEvaluatePreviewEventsResponse,
)
from .new_accounting_sync_configuration_param import (
    NewAccountingSyncConfigurationParam as NewAccountingSyncConfigurationParam,
)
from .subscription_schedule_plan_change_params import (
    SubscriptionSchedulePlanChangeParams as SubscriptionSchedulePlanChangeParams,
)
from .alert_create_for_external_customer_params import (
    AlertCreateForExternalCustomerParams as AlertCreateForExternalCustomerParams,
)
from .new_subscription_grouped_tiered_price_param import (
    NewSubscriptionGroupedTieredPriceParam as NewSubscriptionGroupedTieredPriceParam,
)
from .new_subscription_tiered_package_price_param import (
    NewSubscriptionTieredPackagePriceParam as NewSubscriptionTieredPackagePriceParam,
)
from .subscription_update_fixed_fee_quantity_params import (
    SubscriptionUpdateFixedFeeQuantityParams as SubscriptionUpdateFixedFeeQuantityParams,
)
from .new_subscription_minimum_composite_price_param import (
    NewSubscriptionMinimumCompositePriceParam as NewSubscriptionMinimumCompositePriceParam,
)
from .new_subscription_unit_with_percent_price_param import (
    NewSubscriptionUnitWithPercentPriceParam as NewSubscriptionUnitWithPercentPriceParam,
)
from .new_subscription_grouped_allocation_price_param import (
    NewSubscriptionGroupedAllocationPriceParam as NewSubscriptionGroupedAllocationPriceParam,
)
from .new_subscription_bulk_with_proration_price_param import (
    NewSubscriptionBulkWithProrationPriceParam as NewSubscriptionBulkWithProrationPriceParam,
)
from .new_subscription_tiered_with_minimum_price_param import (
    NewSubscriptionTieredWithMinimumPriceParam as NewSubscriptionTieredWithMinimumPriceParam,
)
from .new_subscription_unit_with_proration_price_param import (
    NewSubscriptionUnitWithProrationPriceParam as NewSubscriptionUnitWithProrationPriceParam,
)
from .new_subscription_grouped_tiered_package_price_param import (
    NewSubscriptionGroupedTieredPackagePriceParam as NewSubscriptionGroupedTieredPackagePriceParam,
)
from .new_subscription_matrix_with_allocation_price_param import (
    NewSubscriptionMatrixWithAllocationPriceParam as NewSubscriptionMatrixWithAllocationPriceParam,
)
from .new_subscription_threshold_total_amount_price_param import (
    NewSubscriptionThresholdTotalAmountPriceParam as NewSubscriptionThresholdTotalAmountPriceParam,
)
from .new_subscription_cumulative_grouped_bulk_price_param import (
    NewSubscriptionCumulativeGroupedBulkPriceParam as NewSubscriptionCumulativeGroupedBulkPriceParam,
)
from .new_subscription_package_with_allocation_price_param import (
    NewSubscriptionPackageWithAllocationPriceParam as NewSubscriptionPackageWithAllocationPriceParam,
)
from .new_subscription_matrix_with_display_name_price_param import (
    NewSubscriptionMatrixWithDisplayNamePriceParam as NewSubscriptionMatrixWithDisplayNamePriceParam,
)
from .new_subscription_max_group_tiered_package_price_param import (
    NewSubscriptionMaxGroupTieredPackagePriceParam as NewSubscriptionMaxGroupTieredPackagePriceParam,
)
from .new_subscription_tiered_package_with_minimum_price_param import (
    NewSubscriptionTieredPackageWithMinimumPriceParam as NewSubscriptionTieredPackageWithMinimumPriceParam,
)
from .new_subscription_grouped_with_metered_minimum_price_param import (
    NewSubscriptionGroupedWithMeteredMinimumPriceParam as NewSubscriptionGroupedWithMeteredMinimumPriceParam,
)
from .subscription_unschedule_fixed_fee_quantity_updates_params import (
    SubscriptionUnscheduleFixedFeeQuantityUpdatesParams as SubscriptionUnscheduleFixedFeeQuantityUpdatesParams,
)
from .new_subscription_grouped_with_prorated_minimum_price_param import (
    NewSubscriptionGroupedWithProratedMinimumPriceParam as NewSubscriptionGroupedWithProratedMinimumPriceParam,
)
from .new_subscription_scalable_matrix_with_unit_pricing_price_param import (
    NewSubscriptionScalableMatrixWithUnitPricingPriceParam as NewSubscriptionScalableMatrixWithUnitPricingPriceParam,
)
from .new_subscription_scalable_matrix_with_tiered_pricing_price_param import (
    NewSubscriptionScalableMatrixWithTieredPricingPriceParam as NewSubscriptionScalableMatrixWithTieredPricingPriceParam,
)



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional

from ..._models import BaseModel

__all__ = ["UnitConfig"]


class UnitConfig(BaseModel):
    unit_amount: str
    """Rate per unit of usage"""

    scaling_factor: Optional[float] = None
    """Multiplier to scale rated quantity by"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Iterable, Optional
from typing_extensions import Literal, Required, TypedDict

from .threshold_param import ThresholdParam

__all__ = ["AlertCreateForExternalCustomerParams"]


class AlertCreateForExternalCustomerParams(TypedDict, total=False):
    currency: Required[str]
    """The case sensitive currency or custom pricing unit to use for this alert."""

    type: Required[Literal["credit_balance_depleted", "credit_balance_dropped", "credit_balance_recovered"]]
    """The type of alert to create. This must be a valid alert type."""

    thresholds: Optional[Iterable[ThresholdParam]]
    """The thresholds that define the values at which the alert will be triggered."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Iterable, Optional
from typing_extensions import Required, TypedDict

from ..._types import SequenceNotStr
from .matrix_value import MatrixValue

__all__ = ["MatrixConfig"]


class MatrixConfig(TypedDict, total=False):
    default_unit_amount: Required[str]
    """Default per unit rate for any usage not bucketed into a specified matrix_value"""

    dimensions: Required[SequenceNotStr[Optional[str]]]
    """One or two event property values to evaluate matrix groups by"""

    matrix_values: Required[Iterable[MatrixValue]]
    """Matrix values configuration"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal

from ..._models import BaseModel

__all__ = ["NewBillingCycleConfiguration"]


class NewBillingCycleConfiguration(BaseModel):
    duration: int
    """The duration of the billing period."""

    duration_unit: Literal["day", "month"]
    """The unit of billing period duration."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Optional
from typing_extensions import Required, TypedDict

__all__ = ["ConversionRateTier"]


class ConversionRateTier(TypedDict, total=False):
    first_unit: Required[float]
    """Exclusive tier starting value"""

    unit_amount: Required[str]
    """Amount per unit of overage"""

    last_unit: Optional[float]
    """Inclusive tier ending value. If null, this is treated as the last tier"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Iterable
from typing_extensions import Required, TypedDict

from .conversion_rate_tier import ConversionRateTier

__all__ = ["ConversionRateTieredConfig"]


class ConversionRateTieredConfig(TypedDict, total=False):
    tiers: Required[Iterable[ConversionRateTier]]
    """Tiers for rating based on total usage quantities into the specified tier"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List

from ..._models import BaseModel
from ..shared.aggregated_cost import AggregatedCost

__all__ = ["CostListByExternalIDResponse"]


class CostListByExternalIDResponse(BaseModel):
    data: List[AggregatedCost]
